# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Test deployment command."""

import os
import shutil
from pathlib import Path
from typing import Tuple
from unittest import mock

import yaml

from autonomy.cli import cli
from autonomy.constants import (
    DEFAULT_BUILD_FOLDER,
    OPEN_AEA_IMAGE_NAME,
    TENDERMINT_IMAGE_NAME,
)
from autonomy.deploy.constants import (
    DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA,
    DEPLOYMENT_KEY_DIRECTORY,
    KUBERNETES_AGENT_KEY_NAME,
)

from tests.conftest import (
    ETHEREUM_ENCRYPTED_KEYS,
    ETHEREUM_ENCRYPTION_PASSWORD,
    ROOT_DIR,
)
from tests.test_autonomy.test_cli.base import BaseCliTest


class TestBuildDeployment(BaseCliTest):
    """Test `autonomy deply build deployment` command."""

    cli_options: Tuple[str, ...] = ("deploy", "build", "deployment")
    service_id: str = "valory/oracle_hardhat"

    keys_file: Path

    @classmethod
    def setup(cls) -> None:
        """Setup class."""

        super().setup()

        cls.keys_file = cls.t / "keys.json"

        shutil.copytree(ROOT_DIR / "packages", cls.t / "packages")
        shutil.copy(
            ROOT_DIR / "deployments" / "keys" / "hardhat_keys.json", cls.keys_file
        )

        os.chdir(cls.t)
        cls.cli_runner.invoke(
            cli, ("deploy", "build", "image", "valory/oracle_hardhat", "--dependencies")
        )

    def _build_images(self, version: str = "0.1.0") -> None:
        """Build images."""
        result = self.cli_runner.invoke(
            cli, ("deploy", "build", "image", self.service_id, "--version", version)
        )

        assert result.exit_code == 0, result.output

        result = self.cli_runner.invoke(
            cli,
            (
                "deploy",
                "build",
                "image",
                self.service_id,
                "--version",
                version,
                "--dependencies",
            ),
        )

        assert result.exit_code == 0, result.output

    def test_docker_compose_build(
        self,
    ) -> None:
        """Run tests."""

        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(self.keys_file),
                    "--packages-dir",
                    str(self.t / "packages"),
                    "--o",
                    str(self.t),
                    "--force",
                )
            )

        build_dir = self.t / "abci_build"

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        build_tree = list(map(lambda x: x.name, build_dir.iterdir()))
        assert any(
            [
                child in build_tree
                for child in ["persistent_storage", "nodes", "docker-compose.yaml"]
            ]
        )

        docker_compose_file = build_dir / "docker-compose.yaml"
        with open(docker_compose_file, "r", encoding="utf-8") as fp:
            docker_compose = yaml.safe_load(fp)

        assert any(
            [key in docker_compose for key in ["version", "services", "networks"]]
        )

        assert any(
            [
                service in docker_compose["services"]
                for service in [
                    *map(lambda i: f"abci{i}", range(4)),
                    *map(lambda i: f"node0{i}", range(4)),
                ]
            ]
        )

    def test_kubernetes_build(
        self,
    ) -> None:
        """Run tests."""

        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(self.keys_file),
                    "--packages-dir",
                    str(self.t / "packages"),
                    "--o",
                    str(self.t),
                    "--kubernetes",
                    "--force",
                )
            )

        build_dir = self.t / "abci_build"

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        build_tree = list(map(lambda x: x.name, build_dir.iterdir()))
        assert any(
            [child in build_tree for child in ["persistent_storage", "build.yaml"]]
        )

    def test_versioning_docker_compose(
        self,
    ) -> None:
        """Run tests."""

        version = "1.0.0"
        self._build_images(version)
        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(self.keys_file),
                    "--packages-dir",
                    str(self.t / "packages"),
                    "--o",
                    str(self.t),
                    "--force",
                    "--version",
                    version,
                )
            )

        assert result.exit_code == 0, result.output

        build_dir = self.t / "abci_build"
        docker_compose_file = build_dir / "docker-compose.yaml"
        with open(docker_compose_file, "r", encoding="utf-8") as fp:
            docker_compose = yaml.safe_load(fp)

        for i in range(4):
            assert (
                docker_compose["services"][f"node{i}"]["image"]
                == f"{TENDERMINT_IMAGE_NAME}:{version}"
            )
            assert (
                docker_compose["services"][f"abci{i}"]["image"]
                == f"{OPEN_AEA_IMAGE_NAME}:oracle_deployable-{version}"
            )

    def test_versioning_kubernetes(
        self,
    ) -> None:
        """Run tests."""

        version = "1.0.0"
        self._build_images(version)
        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(self.keys_file),
                    "--packages-dir",
                    str(self.t / "packages"),
                    "--o",
                    str(self.t),
                    "--force",
                    "--kubernetes",
                    "--version",
                    version,
                )
            )

        assert result.exit_code == 0, result.output

        build_dir = self.t / "abci_build"
        kubernetes_config_file = build_dir / "build.yaml"
        with open(kubernetes_config_file, "r", encoding="utf-8") as fp:
            kubernetes_config = list(yaml.safe_load_all(fp))

        for resource in kubernetes_config:
            try:
                resource["spec"]["template"]["spec"]["containers"][0]["image"]
                resource["spec"]["template"]["spec"]["containers"][1]["image"]
            except (KeyError, IndexError):
                continue

            assert (
                resource["spec"]["template"]["spec"]["containers"][0]["image"]
                == f"{TENDERMINT_IMAGE_NAME}:{version}"
            )

            assert (
                resource["spec"]["template"]["spec"]["containers"][1]["image"]
                == f"{OPEN_AEA_IMAGE_NAME}:oracle_deployable-{version}"
            )

    def test_docker_compose_no_password(
        self,
    ) -> None:
        """Run tests."""

        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(self.keys_file),
                    "--o",
                    str(self.t),
                    "--force",
                )
            )

        build_dir = self.t / DEFAULT_BUILD_FOLDER

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        docker_compose_file = build_dir / "docker-compose.yaml"
        with open(docker_compose_file, "r", encoding="utf-8") as fp:
            docker_compose = yaml.safe_load(fp)

        agents = int(len(docker_compose["services"]) / 2)
        assert all(
            [
                (
                    build_dir
                    / DEPLOYMENT_KEY_DIRECTORY
                    / DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA.format(agent_n=i)
                ).exists()
                for i in range(agents)
            ]
        )
        for x in range(agents):
            env = dict(
                [
                    f.split("=")
                    for f in docker_compose["services"][f"abci{x}"]["environment"]
                ]
            )
            assert "AEA_PASSWORD" not in env.keys()

    def test_docker_compose_password(
        self,
    ) -> None:
        """Run tests."""
        keys_file = Path(ETHEREUM_ENCRYPTED_KEYS)

        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(keys_file),
                    "--o",
                    str(self.t),
                    "--force",
                    "--password",
                    ETHEREUM_ENCRYPTION_PASSWORD,
                )
            )

        build_dir = self.t / DEFAULT_BUILD_FOLDER

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        build_dir = self.t / DEFAULT_BUILD_FOLDER

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        docker_compose_file = build_dir / "docker-compose.yaml"
        with open(docker_compose_file, "r", encoding="utf-8") as fp:
            docker_compose = yaml.safe_load(fp)

        agents = int(len(docker_compose["services"]) / 2)
        assert all(
            [
                (
                    build_dir
                    / DEPLOYMENT_KEY_DIRECTORY
                    / DEPLOYMENT_AGENT_KEY_DIRECTORY_SCHEMA.format(agent_n=i)
                ).exists()
                for i in range(agents)
            ]
        )
        for x in range(agents):
            env = dict(
                [
                    f.split("=")
                    for f in docker_compose["services"][f"abci{x}"]["environment"]
                ]
            )
            assert "AEA_PASSWORD" in env.keys()
            assert env["AEA_PASSWORD"] == ETHEREUM_ENCRYPTION_PASSWORD

    def test_kubernetes_build_password(
        self,
    ) -> None:
        """Run tests."""
        keys_file = Path(ETHEREUM_ENCRYPTED_KEYS)

        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(keys_file),
                    "--o",
                    str(self.t),
                    "--force",
                    "--kubernetes",
                    "--password",
                    ETHEREUM_ENCRYPTION_PASSWORD,
                )
            )

        build_dir = self.t / DEFAULT_BUILD_FOLDER
        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        build_dir = self.t / DEFAULT_BUILD_FOLDER
        kubernetes_config_file = build_dir / "build.yaml"

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()
        with open(kubernetes_config_file, "r", encoding="utf-8") as fp:
            kubernetes_config = list(yaml.safe_load_all(fp))

        for resource in kubernetes_config:
            try:
                agent_vars = {
                    f["name"]: f["value"]
                    for f in resource["spec"]["template"]["spec"]["containers"][1][
                        "env"
                    ]
                }
            except (KeyError, IndexError):
                continue

            assert agent_vars["AEA_PASSWORD"] == ETHEREUM_ENCRYPTION_PASSWORD

        assert all(
            [
                (
                    build_dir
                    / DEPLOYMENT_KEY_DIRECTORY
                    / KUBERNETES_AGENT_KEY_NAME.format(agent_n=i)
                ).exists()
                for i in range(4)
            ]
        )

    def test_kubernetes_build_no_password(
        self,
    ) -> None:
        """Run tests."""

        with mock.patch("os.chown"):
            result = self.run_cli(
                (
                    self.service_id,
                    str(self.keys_file),
                    "--o",
                    str(self.t),
                    "--force",
                    "--kubernetes",
                )
            )

        build_dir = self.t / DEFAULT_BUILD_FOLDER
        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()

        build_dir = self.t / DEFAULT_BUILD_FOLDER
        kubernetes_config_file = build_dir / "build.yaml"

        assert result.exit_code == 0, f"{result.stdout_bytes}\n{result.stderr_bytes}"
        assert build_dir.exists()
        with open(kubernetes_config_file, "r", encoding="utf-8") as fp:
            kubernetes_config = list(yaml.safe_load_all(fp))

        for resource in kubernetes_config:
            try:
                agent_vars = {
                    f["name"]: f["value"]
                    for f in resource["spec"]["template"]["spec"]["containers"][1][
                        "env"
                    ]
                }
            except (KeyError, IndexError):
                continue

            assert "AEA_PASSWORD" not in agent_vars.keys()

        assert all(
            [
                (
                    build_dir
                    / DEPLOYMENT_KEY_DIRECTORY
                    / KUBERNETES_AGENT_KEY_NAME.format(agent_n=i)
                ).exists()
                for i in range(4)
            ]
        )

    @classmethod
    def teardown(cls) -> None:
        """Teardown method."""

        os.chdir(cls.cwd)
        super().teardown()



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for rounds.py file in valory/liquidity_rebalancing_abci."""

import json
from types import MappingProxyType
from typing import Dict, FrozenSet, Mapping, Optional  # noqa : F401
from unittest import mock

from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.liquidity_rebalancing_abci.payloads import (
    SleepPayload,
    StrategyEvaluationPayload,
)
from packages.valory.skills.liquidity_rebalancing_abci.rounds import (  # noqa: F401
    Event,
    SleepRound,
    StrategyEvaluationRound,
    SynchronizedData,
    TransactionHashBaseRound,
)
from packages.valory.skills.price_estimation_abci.payloads import TransactionHashPayload
from packages.valory.skills.transaction_settlement_abci.payloads import ValidatePayload

from tests.test_skills.test_abstract_round_abci.test_base_rounds import (
    BaseCollectSameUntilThresholdRoundTest,
)
from tests.test_skills.test_transaction_settlement_abci.test_rounds import (
    get_participant_to_signature,
)


MAX_PARTICIPANTS: int = 4


def get_participants() -> FrozenSet[str]:
    """Returns test value for participants"""
    return frozenset([f"agent_{i}" for i in range(MAX_PARTICIPANTS)])


def get_participant_to_strategy(
    participants: FrozenSet[str],
) -> Mapping[str, StrategyEvaluationPayload]:
    """Returns test value for participant_to_strategy"""
    return dict(
        [
            (participant, StrategyEvaluationPayload(sender=participant, strategy="{}"))
            for participant in participants
        ]
    )


def get_participant_to_transfers_result(
    participants: FrozenSet[str],
) -> Dict[str, ValidatePayload]:
    """Get participant_to_lp_result"""
    return {
        participant: ValidatePayload(sender=participant) for participant in participants
    }


def get_participant_to_tx_hash(
    participants: FrozenSet[str],
    hash_: Optional[str] = "tx_hash",
    data_: Optional[str] = "tx_data",
) -> Mapping[str, TransactionHashPayload]:
    """participant_to_tx_hash"""
    return {
        participant: TransactionHashPayload(
            sender=participant, tx_hash=json.dumps({"tx_hash": hash_, "tx_data": data_})
        )
        for participant in participants
    }


def get_participant_to_sleep(
    participants: FrozenSet[str],
) -> Mapping[str, SleepPayload]:
    """Returns test value for participant_to_sleep"""
    return dict(
        [
            (participant, SleepPayload(sender=participant))
            for participant in participants
        ]
    )


class TestTransactionHashBaseRound(BaseCollectSameUntilThresholdRoundTest):
    """Test TransactionHashBaseRound"""

    _synchronized_data_class = SynchronizedData
    _event_class = Event

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        test_round = TransactionHashBaseRound(
            self.synchronized_data, self.consensus_params
        )
        payloads = get_participant_to_tx_hash(self.participants)
        self._complete_run(
            self._test_round(
                test_round=test_round,
                round_payloads=payloads,
                synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                    participant_to_tx_hash=MappingProxyType(payloads),
                    most_voted_tx_hash=json.loads(_test_round.most_voted_payload)[
                        "tx_hash"
                    ],
                ),
                synchronized_data_attr_checks=[
                    lambda _synchronized_data: _synchronized_data.participant_to_tx_hash.keys(),
                ],
                most_voted_payload=payloads["agent_1"].tx_hash,
                exit_event=Event.DONE,
            )
        )


class TestStrategyEvaluationRound(BaseCollectSameUntilThresholdRoundTest):
    """Test StrategyEvaluationRound"""

    _synchronized_data_class = SynchronizedData
    _event_class = Event

    def test_run_wait(
        self,
    ) -> None:
        """Run tests."""
        test_round = StrategyEvaluationRound(
            self.synchronized_data, self.consensus_params
        )

        strategy_mock = mock.PropertyMock(return_value=json.dumps(dict(action="wait")))

        with mock.patch.object(
            StrategyEvaluationPayload,
            "strategy",
            strategy_mock,
        ):
            self._complete_run(
                self._test_round(
                    test_round=test_round,
                    round_payloads=get_participant_to_strategy(self.participants),
                    synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                        participant_to_strategy=get_participant_to_strategy(
                            self.participants
                        ),
                        most_voted_strategy=test_round.most_voted_payload,
                    ),
                    synchronized_data_attr_checks=[
                        lambda _synchronized_data: _synchronized_data.participant_to_strategy.keys(),
                    ],
                    most_voted_payload=StrategyEvaluationPayload.strategy,
                    exit_event=Event.DONE,
                )
            )

    def test_run_enter(
        self,
    ) -> None:
        """Run tests."""
        test_round = StrategyEvaluationRound(
            self.synchronized_data, self.consensus_params
        )

        strategy_mock = mock.PropertyMock(return_value=json.dumps(dict(action="enter")))

        with mock.patch.object(
            StrategyEvaluationPayload,
            "strategy",
            strategy_mock,
        ):
            self._complete_run(
                self._test_round(
                    test_round=test_round,
                    round_payloads=get_participant_to_strategy(self.participants),
                    synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                        participant_to_strategy=get_participant_to_strategy(
                            self.participants
                        ),
                        most_voted_strategy=test_round.most_voted_payload,
                    ),
                    synchronized_data_attr_checks=[
                        lambda _synchronized_data: _synchronized_data.participant_to_strategy.keys(),
                    ],
                    most_voted_payload=StrategyEvaluationPayload.strategy,
                    exit_event=Event.DONE_ENTER,
                )
            )

    def test_run_exit(
        self,
    ) -> None:
        """Run tests."""
        test_round = StrategyEvaluationRound(
            self.synchronized_data, self.consensus_params
        )

        strategy_mock = mock.PropertyMock(return_value=json.dumps(dict(action="exit")))

        with mock.patch.object(
            StrategyEvaluationPayload,
            "strategy",
            strategy_mock,
        ):
            self._complete_run(
                self._test_round(
                    test_round=test_round,
                    round_payloads=get_participant_to_strategy(self.participants),
                    synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                        participant_to_strategy=get_participant_to_strategy(
                            self.participants
                        ),
                        most_voted_strategy=test_round.most_voted_payload,
                    ),
                    synchronized_data_attr_checks=[
                        lambda _synchronized_data: _synchronized_data.participant_to_strategy.keys(),
                    ],
                    most_voted_payload=StrategyEvaluationPayload.strategy,
                    exit_event=Event.DONE_EXIT,
                )
            )

    def test_run_swap_back(
        self,
    ) -> None:
        """Run tests."""
        test_round = StrategyEvaluationRound(
            self.synchronized_data, self.consensus_params
        )

        strategy_mock = mock.PropertyMock(
            return_value=json.dumps(dict(action="swap_back"))
        )

        with mock.patch.object(
            StrategyEvaluationPayload,
            "strategy",
            strategy_mock,
        ):
            self._complete_run(
                self._test_round(
                    test_round=test_round,
                    round_payloads=get_participant_to_strategy(self.participants),
                    synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(
                        participant_to_strategy=get_participant_to_strategy(
                            self.participants
                        ),
                        most_voted_strategy=test_round.most_voted_payload,
                    ),
                    synchronized_data_attr_checks=[
                        lambda _synchronized_data: _synchronized_data.participant_to_strategy.keys(),
                    ],
                    most_voted_payload=StrategyEvaluationPayload.strategy,
                    exit_event=Event.DONE_SWAP_BACK,
                )
            )


class TestSleepRound(BaseCollectSameUntilThresholdRoundTest):
    """Test StrategyEvaluationRound"""

    _synchronized_data_class = SynchronizedData
    _event_class = Event

    def test_run(
        self,
    ) -> None:
        """Run tests."""
        test_round = SleepRound(self.synchronized_data, self.consensus_params)
        with mock.patch.object(SleepPayload, "sleep", return_value="sleep"):
            self._complete_run(
                self._test_round(
                    test_round=test_round,
                    round_payloads=get_participant_to_sleep(self.participants),
                    synchronized_data_update_fn=lambda _synchronized_data, _test_round: _synchronized_data.update(),
                    synchronized_data_attr_checks=[],
                    most_voted_payload=SleepPayload.sleep,
                    exit_event=Event.DONE,
                )
            )


def test_synchronized_data() -> None:
    """Test SynchronizedData."""

    participants = get_participants()
    period_count = 0
    setup_params: Dict = {}
    most_voted_strategy: Dict = {}
    most_voted_keeper_address = "0x_keeper"
    safe_contract_address = "0x_contract"
    multisend_contract_address = "0x_multisend"
    most_voted_tx_hash = "tx_hash"
    final_tx_hash = "tx_hash"
    participant_to_lp_result = get_participant_to_transfers_result(participants)
    participant_to_tx_hash = get_participant_to_tx_hash(participants)
    participant_to_signature = get_participant_to_signature(participants)
    participant_to_strategy = get_participant_to_strategy(participants)

    synchronized_data = SynchronizedData(
        AbciAppDB(
            setup_data=AbciAppDB.data_to_lists(
                dict(
                    participants=participants,
                    setup_params=setup_params,
                    most_voted_strategy=most_voted_strategy,
                    most_voted_keeper_address=most_voted_keeper_address,
                    safe_contract_address=safe_contract_address,
                    multisend_contract_address=multisend_contract_address,
                    most_voted_tx_hash=most_voted_tx_hash,
                    final_tx_hash=final_tx_hash,
                    participant_to_lp_result=participant_to_lp_result,
                    participant_to_tx_hash=participant_to_tx_hash,
                    participant_to_signature=participant_to_signature,
                    participant_to_strategy=participant_to_strategy,
                )
            ),
        )
    )

    assert synchronized_data.participants == participants
    assert synchronized_data.period_count == period_count
    assert synchronized_data.most_voted_strategy == most_voted_strategy
    assert synchronized_data.most_voted_keeper_address == most_voted_keeper_address
    assert synchronized_data.safe_contract_address == safe_contract_address
    assert synchronized_data.multisend_contract_address == multisend_contract_address
    assert synchronized_data.most_voted_tx_hash == most_voted_tx_hash
    assert synchronized_data.final_tx_hash == final_tx_hash
    assert synchronized_data.participant_to_tx_hash == participant_to_tx_hash
    assert synchronized_data.participant_to_signature == participant_to_signature
    assert synchronized_data.participant_to_strategy == participant_to_strategy



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tools for aggregating benchmark results."""

import json
from pathlib import Path
from typing import Any, Dict, List

import pandas as pd

from autonomy.analyse.benchmark.html import HTML_TEMPLATE, TABLE_TEMPLATE


class BlockTypes:  # pylint: disable=too-few-public-methods
    """Block types."""

    ALL = "all"
    LOCAL = "local"
    CONSENSUS = "consensus"
    TOTAL = "total"

    types = ("local", "consensus", "total")


def read_benchmark_data(path: Path) -> List[Dict]:
    """Returns logs."""
    benchmark_data = []
    for agent_dir in path.iterdir():
        agent_benchmark_data: Dict[str, Any] = {}
        agent_benchmark_data["name"] = agent_dir.name
        agent_benchmark_data["data"] = dict(
            map(
                lambda path: (
                    int(path.name.replace(".json", "")),
                    json.loads(path.read_text()),
                ),
                agent_dir.glob("*.json"),
            )
        )
        benchmark_data.append(agent_benchmark_data)
    return benchmark_data


def create_dataframe(data: List[Dict]) -> pd.DataFrame:
    """Create pandas.DataFrame object from benchmark data."""

    rows = []
    behaviours = [behaviour_data["behaviour"] for behaviour_data in data[0]["data"][0]]
    cols = ["agent", "period", "block_type", *behaviours]

    for agent in data:
        for period_n, period_data in agent["data"].items():
            for block_t in BlockTypes.types:
                rows.append(
                    {
                        "agent": agent["name"],
                        "period": period_n,
                        "block_type": block_t,
                        **{
                            behaviour_data["behaviour"]: behaviour_data["data"].get(
                                block_t, 0.0
                            )
                            for behaviour_data in period_data
                        },
                    }
                )

    return pd.DataFrame(
        data=rows,
    )[cols]


def format_output(df: pd.DataFrame, period: int, block_type: str) -> str:
    """Format output from given dataframe and parameters"""

    df = df.copy(deep=True).fillna(value=0.0)
    df = df[df["period"] == period]
    df = df[df["block_type"] == block_type]

    del df["period"]
    del df["block_type"]

    time_df = df[df.columns[1:]]
    stats_df = pd.DataFrame(
        data=[
            ["mean", *time_df.mean().values],
            ["median", *time_df.median().values],
            ["std_dev", *time_df.std().values],
        ],
        columns=df.columns,
    )

    output_df = pd.concat((df, stats_df))
    return output_df.to_html(index=False, justify="left")


def aggregate(path: Path, block_type: str, period: int, output: Path) -> None:
    """Benchmark Aggregator."""

    path = Path(path)
    benchmark_data = read_benchmark_data(path)
    dataframe = create_dataframe(benchmark_data)

    periods = range(dataframe.period.max()) if period == -1 else (period,)
    blocks = BlockTypes.types if block_type == BlockTypes.ALL else (block_type,)

    tables = ""
    for period_n in periods:
        for block_t in blocks:
            tables += TABLE_TEMPLATE.format(
                table=format_output(dataframe, period_n, block_t),
                period=period_n,
                block=block_t,
            )

    output.write_text(HTML_TEMPLATE.format(tables=tables), encoding="utf-8")



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/counter_client skill's behaviours."""

import json
from pathlib import Path
from typing import Any, cast

from aea.test_tools.test_skill import BaseSkillTestCase

from packages.valory.connections.http_client.connection import (
    PUBLIC_ID as HTTP_CLIENT_PUBLIC_ID,
)
from packages.valory.skills.counter_client.behaviours import BaseBehaviour
from packages.valory.skills.counter_client.handlers import HttpHandler, HttpMessage

from tests.conftest import ROOT_DIR


class BaseTestClass(BaseSkillTestCase):
    """Base test class."""

    path_to_skill = Path(ROOT_DIR, "packages", "valory", "skills", "counter_client")

    behaviour: BaseBehaviour
    behaviour_name: str
    http_handler: HttpHandler

    @classmethod
    def setup(cls, **kwargs: Any) -> None:
        """Setup the test class."""
        super().setup()
        assert cls._skill.skill_context._agent_context is not None
        cls._skill.skill_context._agent_context.identity._default_address_key = (
            "ethereum"
        )
        cls._skill.skill_context._agent_context._default_ledger_id = "ethereum"
        cls.behaviour = getattr(cls._skill.skill_context.behaviours, cls.behaviour_name)

        cls.http_handler = cast(
            HttpHandler, cls._skill.skill_context.handlers.http_handler
        )


class TestIncrementerBehaviour(BaseTestClass):
    """Test IncrementerBehaviour."""

    behaviour_name = "incrementer"

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        self.behaviour.act_wrapper()
        self.assert_quantity_in_outbox(1)

        actual_message = self.get_message_from_outbox()
        assert actual_message is not None, "No message in outbox."
        has_attr, msg = self.message_has_attributes(
            actual_message=actual_message,
            message_type=HttpMessage,
            to=str(HTTP_CLIENT_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            method="GET",
            headers="",
            url=self.behaviour.tendermint_url + "/broadcast_tx_commit?tx=0x01",
        )

        assert has_attr, msg
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=(actual_message.dialogue_reference[0], "stub"),
            performative=HttpMessage.Performative.RESPONSE,
            target=actual_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(HTTP_CLIENT_PUBLIC_ID),
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=json.dumps({}).encode("utf-8"),
        )
        self.http_handler.handle(incoming_message)
        self.assert_quantity_in_outbox(0)


class TestMonitorBehaviour(BaseTestClass):
    """Test MonitorBehaviour."""

    behaviour_name = "monitor"

    def test_run(
        self,
    ) -> None:
        """Run tests."""

        self.behaviour.act_wrapper()
        self.assert_quantity_in_outbox(1)

        actual_message = self.get_message_from_outbox()
        assert actual_message is not None, "No message in outbox."
        has_attr, msg = self.message_has_attributes(
            actual_message=actual_message,
            message_type=HttpMessage,
            to=str(HTTP_CLIENT_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            method="GET",
            headers="",
            url=self.behaviour.tendermint_url + "/abci_query",
        )

        assert has_attr, msg
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=(actual_message.dialogue_reference[0], "stub"),
            performative=HttpMessage.Performative.RESPONSE,
            target=actual_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(HTTP_CLIENT_PUBLIC_ID),
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=json.dumps({"result": {"response": {"value": "AAAACg=="}}}).encode(
                "utf-8"
            ),
        )
        self.http_handler.handle(incoming_message)
        self.assert_quantity_in_outbox(0)


class TestHttpHandler(BaseTestClass):
    """Test `handle` method from HttpHandle."""

    behaviour_name = "monitor"

    def test_handle(
        self,
    ) -> None:
        """Run tests."""

        self.behaviour.act_wrapper()
        self.assert_quantity_in_outbox(1)

        actual_message = self.get_message_from_outbox()
        assert actual_message is not None, "No message in outbox."
        has_attr, msg = self.message_has_attributes(
            actual_message=actual_message,
            message_type=HttpMessage,
            to=str(HTTP_CLIENT_PUBLIC_ID),
            sender=str(self.skill.skill_context.skill_id),
            method="GET",
            headers="",
            url=self.behaviour.tendermint_url + "/abci_query",
        )

        assert has_attr, msg
        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=("stub", "stub"),
            performative=HttpMessage.Performative.RESPONSE,
            target=actual_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(HTTP_CLIENT_PUBLIC_ID),
            version="",
            status_code=200,
            status_text="",
            headers="",
            body=json.dumps({"result": {"response": {"value": "AAAACg=="}}}).encode(
                "utf-8"
            ),
        )
        self.http_handler.handle(incoming_message)

        incoming_message = self.build_incoming_message(
            message_type=HttpMessage,
            dialogue_reference=(actual_message.dialogue_reference[0], "stub"),
            performative=HttpMessage.Performative.RESPONSE,
            target=actual_message.message_id,
            message_id=-1,
            to=str(self.skill.skill_context.skill_id),
            sender=str(HTTP_CLIENT_PUBLIC_ID),
            version="",
            status_code=201,
            status_text="",
            headers="",
            body=json.dumps({"result": {"response": {"value": "AAAACg=="}}}).encode(
                "utf-8"
            ),
        )
        self.http_handler.handle(incoming_message)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the data classes for the `transaction settlement` ABCI application."""
import textwrap
from abc import ABC
from collections import deque
from enum import Enum
from typing import Deque, Dict, List, Mapping, Optional, Set, Tuple, Type, cast

from packages.valory.skills.abstract_round_abci.base import (
    ABCIAppInternalError,
    AbciApp,
    AbciAppDB,
    AbciAppTransitionFunction,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    CollectDifferentUntilThresholdRound,
    CollectNonEmptyUntilThresholdRound,
    CollectSameUntilThresholdRound,
    DegenerateRound,
    OnlyKeeperSendsRound,
    VotingRound,
)
from packages.valory.skills.transaction_settlement_abci.payload_tools import (
    VerificationStatus,
    tx_hist_hex_to_payload,
)
from packages.valory.skills.transaction_settlement_abci.payloads import (
    CheckTransactionHistoryPayload,
    FinalizationTxPayload,
    RandomnessPayload,
    ResetPayload,
    SelectKeeperPayload,
    SignaturePayload,
    SynchronizeLateMessagesPayload,
    ValidatePayload,
)


ADDRESS_LENGTH = 42
TX_HASH_LENGTH = 66
RETRIES_LENGTH = 64


class Event(Enum):
    """Event enumeration for the price estimation demo."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"
    NEGATIVE = "negative"
    NONE = "none"
    FINALIZE_TIMEOUT = "finalize_timeout"
    VALIDATE_TIMEOUT = "validate_timeout"
    CHECK_TIMEOUT = "check_timeout"
    RESET_TIMEOUT = "reset_timeout"
    CHECK_HISTORY = "check_history"
    CHECK_LATE_ARRIVING_MESSAGE = "check_late_arriving_message"
    FINALIZATION_FAILED = "finalization_failed"
    MISSED_AND_LATE_MESSAGES_MISMATCH = "missed_and_late_messages_mismatch"
    INSUFFICIENT_FUNDS = "insufficient_funds"
    INCORRECT_SERIALIZATION = "incorrect_serialization"


class SynchronizedData(
    BaseSynchronizedData
):  # pylint: disable=too-many-instance-attributes
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    @property
    def safe_contract_address(self) -> str:
        """Get the safe contract address."""
        return cast(str, self.db.get_strict("safe_contract_address"))

    @property
    def participant_to_signature(self) -> Mapping[str, SignaturePayload]:
        """Get the participant_to_signature."""
        return cast(
            Mapping[str, SignaturePayload],
            self.db.get_strict("participant_to_signature"),
        )

    @property
    def tx_hashes_history(self) -> List[str]:
        """Get the current cycle's tx hashes history, which has not yet been verified."""
        raw = cast(str, self.db.get("tx_hashes_history", ""))
        return textwrap.wrap(raw, TX_HASH_LENGTH)

    @property
    def keepers(self) -> Deque[str]:
        """Get the current cycle's keepers who have tried to submit a transaction."""
        if self.is_keeper_set:
            keepers_unparsed = cast(str, self.db.get_strict("keepers"))
            keepers_parsed = textwrap.wrap(
                keepers_unparsed[RETRIES_LENGTH:], ADDRESS_LENGTH
            )
            return deque(keepers_parsed)
        return deque()

    @property
    def keepers_threshold_exceeded(self) -> bool:
        """Check if the number of selected keepers has exceeded the allowed limit."""
        malicious_threshold = self.nb_participants // 3
        return len(self.keepers) > malicious_threshold

    @property
    def most_voted_keeper_address(self) -> str:
        """Get the first in priority keeper to try to re-submit a transaction."""
        return self.keepers[0]

    @property
    def is_keeper_set(self) -> bool:
        """Check whether keeper is set."""
        return bool(self.db.get("keepers", False))

    @property
    def keeper_retries(self) -> int:
        """Get the number of times the current keeper has retried."""
        if self.is_keeper_set:
            keepers_unparsed = cast(str, self.db.get_strict("keepers"))
            keeper_retries = int.from_bytes(
                bytes.fromhex(keepers_unparsed[:RETRIES_LENGTH]), "big"
            )
            return keeper_retries
        return 0

    @property
    def to_be_validated_tx_hash(self) -> str:
        """
        Get the tx hash which is ready for validation.

        This will always be the last hash in the `tx_hashes_history`,
        due to the way we are inserting the hashes in the array.
        We keep the hashes sorted by the time of their finalization.
        If this property is accessed before the finalization succeeds,
        then it is incorrectly used and raises an internal error.

        :return: the tx hash which is ready for validation.
        """
        if len(self.tx_hashes_history) > 0:
            return self.tx_hashes_history[-1]
        raise ABCIAppInternalError(  # pragma: no cover
            "An Error occurred while trying to get the tx hash for validation: "
            "There are no transaction hashes recorded!"
        )

    @property
    def final_tx_hash(self) -> str:
        """Get the verified tx hash."""
        return cast(str, self.db.get_strict("final_tx_hash"))

    @property
    def final_verification_status(self) -> VerificationStatus:
        """Get the final verification status."""
        return cast(
            VerificationStatus,
            self.db.get("final_verification_status", VerificationStatus.NOT_VERIFIED),
        )

    @property
    def most_voted_tx_hash(self) -> str:
        """Get the most_voted_tx_hash."""
        return cast(str, self.db.get_strict("most_voted_tx_hash"))

    @property
    def missed_messages(self) -> int:
        """Check the number of missed messages."""
        return cast(int, self.db.get("missed_messages", 0))

    @property
    def should_check_late_messages(self) -> bool:
        """Check if we should check for late-arriving messages."""
        return self.missed_messages > 0

    @property
    def late_arriving_tx_hashes(self) -> List[str]:
        """Get the late_arriving_tx_hashes."""
        late_arriving_tx_hashes_unparsed = cast(
            List[str], self.db.get_strict("late_arriving_tx_hashes")
        )
        late_arriving_tx_hashes_parsed = []
        for unparsed_hash in late_arriving_tx_hashes_unparsed:
            if len(unparsed_hash) % TX_HASH_LENGTH != 0:
                # if we cannot parse the hashes, then the developer has serialized them incorrectly.
                raise ABCIAppInternalError(
                    f"Cannot parse late arriving hashes: {unparsed_hash}!"
                )
            parsed_hashes = textwrap.wrap(unparsed_hash, TX_HASH_LENGTH)
            late_arriving_tx_hashes_parsed.extend(parsed_hashes)

        return late_arriving_tx_hashes_parsed


class FailedRound(DegenerateRound, ABC):
    """A round that represents that the period failed"""

    round_id = "failed"


class CollectSignatureRound(CollectDifferentUntilThresholdRound):
    """A round in which agents sign the transaction"""

    round_id = "collect_signature"
    allowed_tx_type = SignaturePayload.transaction_type
    payload_attribute = "signature"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    selection_key = "participant"
    collection_key = "participant_to_signature"


class FinalizationRound(OnlyKeeperSendsRound):
    """A round that represents transaction signing has finished"""

    round_id = "finalization"
    allowed_tx_type = FinalizationTxPayload.transaction_type
    payload_attribute = "tx_data"

    def end_block(
        self,
    ) -> Optional[
        Tuple[BaseSynchronizedData, Enum]
    ]:  # pylint: disable=too-many-return-statements
        """Process the end of the block."""
        if not self.has_keeper_sent_payload:
            return None

        if self.keeper_payload is None:  # pragma: no cover
            return self.synchronized_data, Event.FINALIZATION_FAILED

        verification_status = VerificationStatus(self.keeper_payload["status_value"])
        synchronized_data = cast(
            SynchronizedData,
            self.synchronized_data.update(
                synchronized_data_class=SynchronizedData,
                tx_hashes_history=self.keeper_payload["tx_hashes_history"],
                final_verification_status=verification_status,
                keepers=self.keeper_payload["serialized_keepers"],
                blacklisted_keepers=self.keeper_payload["blacklisted_keepers"],
            ),
        )

        # check if we succeeded in finalization.
        # we may fail in any of the following cases:
        # 1. Getting raw safe transaction.
        # 2. Requesting transaction signature.
        # 3. Requesting transaction digest.
        if self.keeper_payload["received_hash"]:
            return synchronized_data, Event.DONE
        # If keeper has been blacklisted, return an `INSUFFICIENT_FUNDS` event.
        if verification_status == VerificationStatus.INSUFFICIENT_FUNDS:
            return synchronized_data, Event.INSUFFICIENT_FUNDS
        # This means that getting raw safe transaction succeeded,
        # but either requesting tx signature or requesting tx digest failed.
        if verification_status not in (
            VerificationStatus.ERROR,
            VerificationStatus.VERIFIED,
        ):
            return synchronized_data, Event.FINALIZATION_FAILED
        # if there is a tx hash history, then check it for validated txs.
        if synchronized_data.tx_hashes_history:
            return synchronized_data, Event.CHECK_HISTORY
        # if there could be any late messages, check if any has arrived.
        if synchronized_data.should_check_late_messages:
            return synchronized_data, Event.CHECK_LATE_ARRIVING_MESSAGE
        # otherwise fail.
        return synchronized_data, Event.FINALIZATION_FAILED


class RandomnessTransactionSubmissionRound(CollectSameUntilThresholdRound):
    """A round for generating randomness"""

    round_id = "randomness_transaction_submission"
    allowed_tx_type = RandomnessPayload.transaction_type
    payload_attribute = "randomness"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_randomness"
    selection_key = "most_voted_randomness"


class SelectKeeperTransactionSubmissionRoundA(CollectSameUntilThresholdRound):
    """A round in which a keeper is selected for transaction submission"""

    round_id = "select_keeper_transaction_submission_a"
    allowed_tx_type = SelectKeeperPayload.transaction_type
    payload_attribute = "keepers"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_selection"
    selection_key = "keepers"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Process the end of the block."""

        if self.threshold_reached and self.most_voted_payload is not None:
            if (
                len(self.most_voted_payload) < RETRIES_LENGTH + ADDRESS_LENGTH
                or (len(self.most_voted_payload) - RETRIES_LENGTH) % ADDRESS_LENGTH != 0
            ):
                # if we cannot parse the keepers' payload, then the developer has serialized it incorrectly.
                return self.synchronized_data, Event.INCORRECT_SERIALIZATION

        return super().end_block()


class SelectKeeperTransactionSubmissionRoundB(SelectKeeperTransactionSubmissionRoundA):
    """A round in which a new keeper is selected for transaction submission"""

    round_id = "select_keeper_transaction_submission_b"


class SelectKeeperTransactionSubmissionRoundBAfterTimeout(
    SelectKeeperTransactionSubmissionRoundB
):
    """A round in which a new keeper is selected for tx submission after a round timeout of the previous keeper"""

    round_id = "select_keeper_transaction_submission_b_after_timeout"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Process the end of the block."""
        if self.threshold_reached:
            synchronized_data = cast(
                SynchronizedData,
                self.synchronized_data.update(
                    missed_messages=cast(
                        SynchronizedData, self.synchronized_data
                    ).missed_messages
                    + 1
                ),
            )
            if synchronized_data.keepers_threshold_exceeded:
                # we only stop re-selection if there are any previous transaction hashes or any missed messages.
                if len(synchronized_data.tx_hashes_history) > 0:
                    return synchronized_data, Event.CHECK_HISTORY
                if synchronized_data.should_check_late_messages:
                    return synchronized_data, Event.CHECK_LATE_ARRIVING_MESSAGE
        return super().end_block()


class ValidateTransactionRound(VotingRound):
    """A round in which agents validate the transaction"""

    round_id = "validate_transaction"
    allowed_tx_type = ValidatePayload.transaction_type
    payload_attribute = "vote"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    negative_event = Event.NEGATIVE
    none_event = Event.NONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_votes"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Process the end of the block."""
        # if reached participant threshold, set the result
        if self.positive_vote_threshold_reached:
            # We only set the final tx hash if we are about to exit from the transaction settlement skill.
            # Then, the skills which use the transaction settlement can check the tx hash
            # and if it is None, then it means that the transaction has failed.
            synchronized_data = self.synchronized_data.update(
                synchronized_data_class=self.synchronized_data_class,
                participant_to_votes=self.collection,
                final_verification_status=VerificationStatus.VERIFIED,
                final_tx_hash=cast(
                    SynchronizedData, self.synchronized_data
                ).to_be_validated_tx_hash,
            )  # type: ignore
            return synchronized_data, self.done_event
        if self.negative_vote_threshold_reached:
            return self.synchronized_data, self.negative_event
        if self.none_vote_threshold_reached:
            return self.synchronized_data, self.none_event
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, self.no_majority_event
        return None


class CheckTransactionHistoryRound(CollectSameUntilThresholdRound):
    """A round in which agents check the transaction history to see if any previous tx has been validated"""

    round_id = "check_transaction_history"
    allowed_tx_type = CheckTransactionHistoryPayload.transaction_type
    payload_attribute = "verified_res"
    synchronized_data_class = SynchronizedData
    selection_key = "most_voted_check_result"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Enum]]:
        """Process the end of the block."""
        if self.threshold_reached:
            return_status, return_tx_hash = tx_hist_hex_to_payload(
                self.most_voted_payload
            )

            if return_status == VerificationStatus.NOT_VERIFIED:
                # We don't update the synchronized_data as we need to repeat all checks again later
                synchronized_data = self.synchronized_data
            else:
                # We only set the final tx hash if we are about to exit from the transaction settlement skill.
                # Then, the skills which use the transaction settlement can check the tx hash
                # and if it is None, then it means that the transaction has failed.
                synchronized_data = self.synchronized_data.update(
                    synchronized_data_class=self.synchronized_data_class,
                    participant_to_check=self.collection,
                    final_verification_status=return_status,
                    final_tx_hash=return_tx_hash,
                )

            if return_status == VerificationStatus.VERIFIED:
                return synchronized_data, Event.DONE
            if (
                return_status == VerificationStatus.NOT_VERIFIED
                and cast(
                    SynchronizedData, self.synchronized_data
                ).should_check_late_messages
            ):
                return synchronized_data, Event.CHECK_LATE_ARRIVING_MESSAGE
            if return_status == VerificationStatus.NOT_VERIFIED:
                return synchronized_data, Event.NEGATIVE

            return synchronized_data, Event.NONE

        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, Event.NO_MAJORITY
        return None


class CheckLateTxHashesRound(CheckTransactionHistoryRound):
    """A round in which agents check the late-arriving transaction hashes to see if any of them has been validated"""

    round_id = "check_late_tx_hashes"


class SynchronizeLateMessagesRound(CollectNonEmptyUntilThresholdRound):
    """A round in which agents synchronize potentially late arriving messages"""

    round_id = "synchronize_late_messages"
    allowed_tx_type = SynchronizeLateMessagesPayload.transaction_type
    payload_attribute = "tx_hashes"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    none_event = Event.NONE
    selection_key = "participant"
    collection_key = "late_arriving_tx_hashes"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        result = super().end_block()
        if result is None:
            return None

        synchronized_data, event = cast(Tuple[BaseSynchronizedData, Event], result)

        synchronized_data = cast(SynchronizedData, self.synchronized_data)
        n_late_arriving_tx_hashes = len(synchronized_data.late_arriving_tx_hashes)
        if n_late_arriving_tx_hashes > synchronized_data.missed_messages:
            return synchronized_data, Event.MISSED_AND_LATE_MESSAGES_MISMATCH

        synchronized_data = synchronized_data.update(
            missed_messages=synchronized_data.missed_messages
            - n_late_arriving_tx_hashes
        )
        return synchronized_data, event


class FinishedTransactionSubmissionRound(DegenerateRound, ABC):
    """A round that represents the transition to the ResetAndPauseRound"""

    round_id = "pre_reset_and_pause"


class ResetRound(CollectSameUntilThresholdRound):
    """A round that represents the reset of a period"""

    round_id = "reset"
    allowed_tx_type = ResetPayload.transaction_type
    payload_attribute = "period_count"

    def end_block(self) -> Optional[Tuple[BaseSynchronizedData, Event]]:
        """Process the end of the block."""
        if self.threshold_reached:
            latest_data = self.synchronized_data.db.get_latest()
            synchronized_data = self.synchronized_data.create(
                synchronized_data_class=self.synchronized_data_class,
                **AbciAppDB.data_to_lists(latest_data),
            )
            return synchronized_data, Event.DONE
        if not self.is_majority_possible(
            self.collection, self.synchronized_data.nb_participants
        ):
            return self.synchronized_data, Event.NO_MAJORITY
        return None


class TransactionSubmissionAbciApp(AbciApp[Event]):
    """TransactionSubmissionAbciApp

    Initial round: RandomnessTransactionSubmissionRound

    Initial states: {RandomnessTransactionSubmissionRound}

    Transition states:
        0. RandomnessTransactionSubmissionRound
            - done: 1.
            - round timeout: 0.
            - no majority: 0.
        1. SelectKeeperTransactionSubmissionRoundA
            - done: 2.
            - round timeout: 1.
            - no majority: 10.
            - incorrect serialization: 12.
        2. CollectSignatureRound
            - done: 3.
            - round timeout: 2.
            - no majority: 10.
        3. FinalizationRound
            - done: 4.
            - check history: 5.
            - finalize timeout: 7.
            - finalization failed: 6.
            - check late arriving message: 8.
            - insufficient funds: 6.
        4. ValidateTransactionRound
            - done: 11.
            - negative: 5.
            - none: 6.
            - validate timeout: 6.
            - no majority: 4.
        5. CheckTransactionHistoryRound
            - done: 11.
            - negative: 6.
            - none: 12.
            - check timeout: 5.
            - no majority: 5.
            - check late arriving message: 8.
        6. SelectKeeperTransactionSubmissionRoundB
            - done: 3.
            - round timeout: 6.
            - no majority: 10.
            - incorrect serialization: 12.
        7. SelectKeeperTransactionSubmissionRoundBAfterTimeout
            - done: 3.
            - check history: 5.
            - check late arriving message: 8.
            - round timeout: 7.
            - no majority: 10.
            - incorrect serialization: 12.
        8. SynchronizeLateMessagesRound
            - done: 9.
            - round timeout: 8.
            - no majority: 8.
            - none: 6.
            - missed and late messages mismatch: 12.
        9. CheckLateTxHashesRound
            - done: 11.
            - negative: 12.
            - none: 12.
            - check timeout: 9.
            - no majority: 12.
            - check late arriving message: 8.
        10. ResetRound
            - done: 0.
            - reset timeout: 12.
            - no majority: 12.
        11. FinishedTransactionSubmissionRound
        12. FailedRound

    Final states: {FailedRound, FinishedTransactionSubmissionRound}

    Timeouts:
        round timeout: 30.0
        finalize timeout: 30.0
        validate timeout: 30.0
        check timeout: 30.0
        reset timeout: 30.0
    """

    initial_round_cls: Type[AbstractRound] = RandomnessTransactionSubmissionRound
    transition_function: AbciAppTransitionFunction = {
        RandomnessTransactionSubmissionRound: {
            Event.DONE: SelectKeeperTransactionSubmissionRoundA,
            Event.ROUND_TIMEOUT: RandomnessTransactionSubmissionRound,
            Event.NO_MAJORITY: RandomnessTransactionSubmissionRound,
        },
        SelectKeeperTransactionSubmissionRoundA: {
            Event.DONE: CollectSignatureRound,
            Event.ROUND_TIMEOUT: SelectKeeperTransactionSubmissionRoundA,
            Event.NO_MAJORITY: ResetRound,
            Event.INCORRECT_SERIALIZATION: FailedRound,
        },
        CollectSignatureRound: {
            Event.DONE: FinalizationRound,
            Event.ROUND_TIMEOUT: CollectSignatureRound,
            Event.NO_MAJORITY: ResetRound,
        },
        FinalizationRound: {
            Event.DONE: ValidateTransactionRound,
            Event.CHECK_HISTORY: CheckTransactionHistoryRound,
            Event.FINALIZE_TIMEOUT: SelectKeeperTransactionSubmissionRoundBAfterTimeout,
            Event.FINALIZATION_FAILED: SelectKeeperTransactionSubmissionRoundB,
            Event.CHECK_LATE_ARRIVING_MESSAGE: SynchronizeLateMessagesRound,
            Event.INSUFFICIENT_FUNDS: SelectKeeperTransactionSubmissionRoundB,
        },
        ValidateTransactionRound: {
            Event.DONE: FinishedTransactionSubmissionRound,
            Event.NEGATIVE: CheckTransactionHistoryRound,
            Event.NONE: SelectKeeperTransactionSubmissionRoundB,
            Event.VALIDATE_TIMEOUT: SelectKeeperTransactionSubmissionRoundB,
            Event.NO_MAJORITY: ValidateTransactionRound,
        },
        CheckTransactionHistoryRound: {
            Event.DONE: FinishedTransactionSubmissionRound,
            Event.NEGATIVE: SelectKeeperTransactionSubmissionRoundB,
            Event.NONE: FailedRound,
            Event.CHECK_TIMEOUT: CheckTransactionHistoryRound,
            Event.NO_MAJORITY: CheckTransactionHistoryRound,
            Event.CHECK_LATE_ARRIVING_MESSAGE: SynchronizeLateMessagesRound,
        },
        SelectKeeperTransactionSubmissionRoundB: {
            Event.DONE: FinalizationRound,
            Event.ROUND_TIMEOUT: SelectKeeperTransactionSubmissionRoundB,
            Event.NO_MAJORITY: ResetRound,
            Event.INCORRECT_SERIALIZATION: FailedRound,
        },
        SelectKeeperTransactionSubmissionRoundBAfterTimeout: {
            Event.DONE: FinalizationRound,
            Event.CHECK_HISTORY: CheckTransactionHistoryRound,
            Event.CHECK_LATE_ARRIVING_MESSAGE: SynchronizeLateMessagesRound,
            Event.ROUND_TIMEOUT: SelectKeeperTransactionSubmissionRoundBAfterTimeout,
            Event.NO_MAJORITY: ResetRound,
            Event.INCORRECT_SERIALIZATION: FailedRound,
        },
        SynchronizeLateMessagesRound: {
            Event.DONE: CheckLateTxHashesRound,
            Event.ROUND_TIMEOUT: SynchronizeLateMessagesRound,
            Event.NO_MAJORITY: SynchronizeLateMessagesRound,
            Event.NONE: SelectKeeperTransactionSubmissionRoundB,
            Event.MISSED_AND_LATE_MESSAGES_MISMATCH: FailedRound,
        },
        CheckLateTxHashesRound: {
            Event.DONE: FinishedTransactionSubmissionRound,
            Event.NEGATIVE: FailedRound,
            Event.NONE: FailedRound,
            Event.CHECK_TIMEOUT: CheckLateTxHashesRound,
            Event.NO_MAJORITY: FailedRound,
            Event.CHECK_LATE_ARRIVING_MESSAGE: SynchronizeLateMessagesRound,
        },
        ResetRound: {
            Event.DONE: RandomnessTransactionSubmissionRound,
            Event.RESET_TIMEOUT: FailedRound,
            Event.NO_MAJORITY: FailedRound,
        },
        FinishedTransactionSubmissionRound: {},
        FailedRound: {},
    }
    final_states: Set[AppState] = {
        FinishedTransactionSubmissionRound,
        FailedRound,
    }
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
        Event.FINALIZE_TIMEOUT: 30.0,
        Event.VALIDATE_TIMEOUT: 30.0,
        Event.CHECK_TIMEOUT: 30.0,
        Event.RESET_TIMEOUT: 30.0,
    }



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for the loading functionality of abstract round abci."""
import os.path
from pathlib import PosixPath
from typing import Dict, Optional, cast

import pandas as pd
import pytest

from packages.valory.skills.abstract_round_abci.io.load import (
    CSVLoader,
    CustomLoaderType,
    ForecasterLoader,
    JSONLoader,
    Loader,
    SupportedLoaderType,
)
from packages.valory.skills.abstract_round_abci.io.store import (
    SupportedFiletype,
    SupportedMultipleObjectsType,
)


class TestLoader:
    """Tests for the `Loader`."""

    def __dummy_custom_loader(self) -> None:
        """A dummy custom loading function to use for the tests."""

    @staticmethod
    @pytest.mark.parametrize(
        "filetype, custom_loader, expected_loader",
        (
            (None, None, None),
            (SupportedFiletype.CSV, None, CSVLoader.load_single_file),
            (SupportedFiletype.JSON, None, JSONLoader.load_single_file),
            (SupportedFiletype.PM_PIPELINE, None, ForecasterLoader.load_single_file),
            (SupportedFiletype.CSV, __dummy_custom_loader, CSVLoader.load_single_file),
            (
                SupportedFiletype.JSON,
                __dummy_custom_loader,
                JSONLoader.load_single_file,
            ),
            (
                SupportedFiletype.PM_PIPELINE,
                __dummy_custom_loader,
                ForecasterLoader.load_single_file,
            ),
            (None, __dummy_custom_loader, __dummy_custom_loader),
        ),
    )
    def test__get_loader_from_filetype(
        filetype: Optional[SupportedFiletype],
        custom_loader: CustomLoaderType,
        expected_loader: Optional[SupportedLoaderType],
    ) -> None:
        """Test `_get_loader_from_filetype`."""
        if all(
            test_arg is None for test_arg in (filetype, custom_loader, expected_loader)
        ):
            with pytest.raises(
                ValueError,
                match="Please provide either a supported filetype or a custom loader function.",
            ):
                Loader(filetype, custom_loader)._get_single_loader_from_filetype()

        else:
            expected_loader = cast(SupportedLoaderType, expected_loader)
            loader = Loader(filetype, custom_loader)
            assert (
                loader._get_single_loader_from_filetype().__code__.co_code
                == expected_loader.__code__.co_code
            )

    @staticmethod
    @pytest.mark.parametrize("multiple", (True, False))
    def test_load(
        multiple: bool,
        tmp_path: PosixPath,
        dummy_obj: pd.DataFrame,
        dummy_multiple_obj: Dict[str, pd.DataFrame],
    ) -> None:
        """Test `load`."""
        loader = Loader(SupportedFiletype.CSV, None)

        if multiple:
            # test with non-directory path
            with pytest.raises(
                ValueError,
                match="Cannot load multiple files from `non_directory_path`! "
                "Please make sure that the path is a folder containing the files.",
            ):
                loader.load("non_directory_path", multiple)

            # create a dummy folder to store multiple dummy objects to test the loader with them.
            dummy_folder_path = os.path.join(tmp_path, "dummy_folder")
            os.mkdir(dummy_folder_path)
            # store dummy objects.
            for name, obj in dummy_multiple_obj.items():
                path = os.path.join(dummy_folder_path, name)
                obj.to_csv(path, index=False)

            # load with loader.
            loaded_objects = cast(
                SupportedMultipleObjectsType, loader.load(dummy_folder_path, multiple)
            )
            assert len(loaded_objects) == len(
                dummy_multiple_obj
            ), "loaded objects length and dummy objects length do not match."
            assert set(loaded_objects.keys()) == set(
                dummy_multiple_obj.keys()
            ), "loaded objects and dummy objects filenames do not match."

            # iterate through the loaded objects and their filenames and the dummy objects and their filenames.
            for actual_frame, expected_frame in zip(
                loaded_objects.values(), dummy_multiple_obj.values()
            ):
                # assert loaded frame with expected.
                pd.testing.assert_frame_equal(actual_frame, expected_frame)

        else:
            # store dummy object.
            path = os.path.join(tmp_path, "dummy_obj.csv")
            dummy_obj.to_csv(path, index=False)
            # load with loader.
            loaded_obj = loader.load(path, multiple)
            # assert loaded frame with expected.
            pd.testing.assert_frame_equal(loaded_obj, dummy_obj)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the data classes for the oracle deployment ABCI application."""

from enum import Enum
from typing import Dict, Set, Type, cast

from packages.valory.skills.abstract_round_abci.base import (
    AbciApp,
    AbciAppTransitionFunction,
    AbstractRound,
    AppState,
    BaseSynchronizedData,
    CollectSameUntilThresholdRound,
    DegenerateRound,
    OnlyKeeperSendsRound,
    VotingRound,
)
from packages.valory.skills.oracle_deployment_abci.payloads import (
    DeployOraclePayload,
    RandomnessPayload,
    SelectKeeperPayload,
    ValidateOraclePayload,
)


class Event(Enum):
    """Event enumeration for the price estimation demo."""

    DONE = "done"
    ROUND_TIMEOUT = "round_timeout"
    NO_MAJORITY = "no_majority"
    NEGATIVE = "negative"
    NONE = "none"
    FAILED = "failed"
    DEPLOY_TIMEOUT = "deploy_timeout"
    VALIDATE_TIMEOUT = "validate_timeout"


class SynchronizedData(BaseSynchronizedData):
    """
    Class to represent the synchronized data.

    This data is replicated by the tendermint application.
    """

    @property
    def safe_contract_address(self) -> str:
        """Get the safe contract address."""
        return cast(str, self.db.get_strict("safe_contract_address"))

    @property
    def oracle_contract_address(self) -> str:
        """Get the oracle contract address."""
        return cast(str, self.db.get("oracle_contract_address"))


class RandomnessOracleRound(CollectSameUntilThresholdRound):
    """A round for generating randomness"""

    round_id = "randomness_oracle"
    allowed_tx_type = RandomnessPayload.transaction_type
    payload_attribute = "randomness"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_randomness"
    selection_key = "most_voted_randomness"


class SelectKeeperOracleRound(CollectSameUntilThresholdRound):
    """A round in a which keeper is selected"""

    round_id = "select_keeper_oracle"
    allowed_tx_type = SelectKeeperPayload.transaction_type
    payload_attribute = "keeper"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_selection"
    selection_key = "most_voted_keeper_address"


class DeployOracleRound(OnlyKeeperSendsRound):
    """A round in a which the oracle is deployed"""

    round_id = "deploy_oracle"
    allowed_tx_type = DeployOraclePayload.transaction_type
    payload_attribute = "oracle_contract_address"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    fail_event = Event.FAILED
    payload_key = "oracle_contract_address"


class ValidateOracleRound(VotingRound):
    """A round in a which the oracle address is validated"""

    round_id = "validate_oracle"
    allowed_tx_type = ValidateOraclePayload.transaction_type
    payload_attribute = "vote"
    synchronized_data_class = SynchronizedData
    done_event = Event.DONE
    negative_event = Event.NEGATIVE
    none_event = Event.NONE
    no_majority_event = Event.NO_MAJORITY
    collection_key = "participant_to_votes"


class FinishedOracleRound(DegenerateRound):
    """A round that represents that the oracle has been deployed"""

    round_id = "finished_oracle"


class OracleDeploymentAbciApp(AbciApp[Event]):
    """OracleDeploymentAbciApp

    Initial round: RandomnessOracleRound

    Initial states: {RandomnessOracleRound}

    Transition states:
        0. RandomnessOracleRound
            - done: 1.
            - round timeout: 0.
            - no majority: 0.
        1. SelectKeeperOracleRound
            - done: 2.
            - round timeout: 0.
            - no majority: 0.
        2. DeployOracleRound
            - done: 3.
            - deploy timeout: 1.
            - failed: 1.
        3. ValidateOracleRound
            - done: 4.
            - negative: 0.
            - none: 0.
            - validate timeout: 0.
            - no majority: 0.
        4. FinishedOracleRound

    Final states: {FinishedOracleRound}

    Timeouts:
        round timeout: 30.0
        validate timeout: 30.0
        deploy timeout: 30.0
    """

    initial_round_cls: Type[AbstractRound] = RandomnessOracleRound
    transition_function: AbciAppTransitionFunction = {
        RandomnessOracleRound: {
            Event.DONE: SelectKeeperOracleRound,
            Event.ROUND_TIMEOUT: RandomnessOracleRound,
            Event.NO_MAJORITY: RandomnessOracleRound,
        },
        SelectKeeperOracleRound: {
            Event.DONE: DeployOracleRound,
            Event.ROUND_TIMEOUT: RandomnessOracleRound,
            Event.NO_MAJORITY: RandomnessOracleRound,
        },
        DeployOracleRound: {
            Event.DONE: ValidateOracleRound,
            Event.DEPLOY_TIMEOUT: SelectKeeperOracleRound,
            # NOTE: Consider the case where the keeper does send the tx but doesn't share the hash.
            # We do not need to check for this! A simple round timeout will do here as the agents
            # do not care about this oracle instance.
            Event.FAILED: SelectKeeperOracleRound,
        },
        ValidateOracleRound: {
            Event.DONE: FinishedOracleRound,
            Event.NEGATIVE: RandomnessOracleRound,
            Event.NONE: RandomnessOracleRound,  # NOTE: unreachable
            Event.VALIDATE_TIMEOUT: RandomnessOracleRound,
            Event.NO_MAJORITY: RandomnessOracleRound,
        },
        FinishedOracleRound: {},
    }
    final_states: Set[AppState] = {FinishedOracleRound}
    event_to_timeout: Dict[Event, float] = {
        Event.ROUND_TIMEOUT: 30.0,
        Event.VALIDATE_TIMEOUT: 30.0,
        Event.DEPLOY_TIMEOUT: 30.0,
    }
    cross_period_persisted_keys = ["oracle_contract_address"]



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Integration tests for the valory/counter skill."""
import base64
import json
import logging
import random
import struct
import time
import warnings
from pathlib import Path

import pytest
import requests
from aea.test_tools.test_cases import AEATestCaseMany

from tests.conftest import ANY_ADDRESS
from tests.fixture_helpers import UseTendermint
from tests.helpers.constants import DEFAULT_REQUESTS_TIMEOUT
from tests.helpers.tendermint_utils import (
    BaseTendermintTestClass,
    TendermintLocalNetworkBuilder,
)


@pytest.mark.e2e
class BaseTestABCICounterSkill:
    """Base test class."""

    def _send_tx_to_node(
        self, node_address: str, value: int, expected_status_code: int
    ) -> None:
        """
        Send a transaction to a certain node.

        :param node_address: the node address
        :param value: the value to propose in the tx
        :param expected_status_code: the expected status code
        """
        # at least two hex digits
        tx_arg = "0x{:02x}".format(value)
        result = requests.get(
            node_address + "/broadcast_tx_commit",
            params=dict(tx=tx_arg),
            timeout=DEFAULT_REQUESTS_TIMEOUT,
        )
        assert result.status_code == expected_status_code

    def _query_node(self, node_address: str, expected_value: int) -> None:
        """
        Send a query to a node about the state of the replicated counter.

        :param node_address: the node address
        :param expected_value: the expected value of the counter.
        """
        result = requests.get(
            node_address + "/abci_query",
            timeout=DEFAULT_REQUESTS_TIMEOUT,
        )
        assert result.status_code == 200
        counter_value_base64 = result.json()["result"]["response"]["value"].encode(
            "ascii"
        )
        counter_value_bytes = base64.b64decode(counter_value_base64)
        counter_value, *_ = struct.unpack(">I", counter_value_bytes)
        assert counter_value == expected_value


@pytest.mark.e2e
@pytest.mark.integration
class TestABCICounterSkill(AEATestCaseMany, UseTendermint):
    """Test that the ABCI counter skill works together with Tendermint."""

    IS_LOCAL = True
    capture_log = True
    cli_log_options = ["-v", "DEBUG"]

    def test_run(self) -> None:
        """Run the ABCI skill."""
        agent_name = "counter_aea"
        self.fetch_agent("valory/counter:0.1.0", agent_name, is_local=self.IS_LOCAL)
        self.set_agent_context(agent_name)
        self.generate_private_key("ethereum")
        self.add_private_key("ethereum", "ethereum_private_key.txt")
        self.set_config("vendor.valory.connections.abci.config.use_tendermint", False)
        self.set_config(
            "vendor.valory.connections.abci.config.host",
            ANY_ADDRESS,
        )
        self.set_config(
            "vendor.valory.connections.abci.config.port",
            self.abci_port,
        )

        process = self.run_agent()
        is_running = self.is_running(process)
        assert is_running, "AEA not running within timeout!"

        time.sleep(2.0)

        check_strings = (
            "ABCI Handler: setup method called.",
            "Received ABCI request of type info",
            "Received ABCI request of type flush",
            "Received ABCI request of type begin_block",
            "Received ABCI request of type commit",
            "Received ABCI request of type end_block",
        )
        missing_strings = self.missing_from_output(process, check_strings)
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in agent output.".format(missing_strings)

        if not self.is_successfully_terminated(process):
            warnings.warn(
                UserWarning(
                    f"ABCI agent with process {process} wasn't successfully terminated."
                )
            )


class TestABCICounterSkillMany(
    AEATestCaseMany, BaseTendermintTestClass, BaseTestABCICounterSkill
):
    """Test that the ABCI counter skill works together with Tendermint."""

    IS_LOCAL = True
    capture_log = True
    NB_AGENTS = 4
    NB_TX = 15

    def test_run(self) -> None:
        """Run the ABCI skill."""
        self.agent_names = [f"agent_{i:05d}" for i in range(self.NB_AGENTS)]
        processes = []
        self.tendermint_net_builder = TendermintLocalNetworkBuilder(
            self.NB_AGENTS, Path(self.t)
        )

        for agent_id, agent_name in enumerate(self.agent_names):
            logging.debug(f"Processing agent {agent_name}...")
            node = self.tendermint_net_builder.nodes[agent_id]
            self.fetch_agent("valory/counter:0.1.0", agent_name, is_local=self.IS_LOCAL)
            self.set_agent_context(agent_name)
            self.generate_private_key("ethereum")
            self.add_private_key("ethereum", "ethereum_private_key.txt")
            # each agent has its Tendermint node instance
            self.set_config(
                "vendor.valory.connections.abci.config.port",
                node.abci_port,
            )
            self.set_config(
                "vendor.valory.connections.abci.config.tendermint_config.rpc_laddr",
                node.rpc_laddr,
            )
            self.set_config(
                "vendor.valory.connections.abci.config.tendermint_config.p2p_laddr",
                node.p2p_laddr,
            )
            self.set_config(
                "vendor.valory.connections.abci.config.tendermint_config.home",
                str(node.home),
            )
            self.set_config(
                "vendor.valory.connections.abci.config.tendermint_config.p2p_seeds",
                json.dumps(self.tendermint_net_builder.get_p2p_seeds()),
                "list",
            )
            process = self.run_agent()
            processes.append(process)

        logging.info("Waiting Tendermint nodes to be up")
        self.health_check(self.tendermint_net_builder)

        # start test of the ABCI app
        # check that the initial counter state is 0
        self._query_agents(0)
        # send transactions to each node, randomly
        for tx_number in range(self.NB_TX):
            time.sleep(0.5)
            agent_id = random.randint(0, self.NB_AGENTS - 1)  # nosec
            new_value = tx_number + 1
            self._send_tx(new_value, agent_id, 200)
        # wait synchronization
        time.sleep(2.0)
        # check that the final counter state is N
        self._query_agents(self.NB_TX)
        # end test of the ABCI app

        check_strings = (*[f"The new count is: {x + 1}" for x in range(self.NB_TX)],)

        # check that *each* AEA prints these messages
        for process in processes:
            missing_strings = self.missing_from_output(process, check_strings)
            assert (
                missing_strings == []
            ), "Strings {} didn't appear in agent output.".format(missing_strings)

            if not self.is_successfully_terminated(process):
                warnings.warn(
                    UserWarning(
                        f"ABCI agent with process {process} wasn't successfully terminated."
                    )
                )

    def _query_agents(self, expected_value: int) -> None:
        """
        Send queries to each agent about the state of the replicated counter.

        :param expected_value: the expected value of the counter.
        """
        for agent_id in range(self.NB_AGENTS):
            self._query_agent(expected_value, agent_id)

    def _send_tx(self, value: int, agent_id: int, expected_status_code: int) -> None:
        """
        Send a transaction to a certain agent.

        :param value: the value to propose in the tx
        :param agent_id: the agent id
        :param expected_status_code: the expected status code
        """
        node = self.tendermint_net_builder.nodes[agent_id]
        self._send_tx_to_node(
            node.get_http_addr("localhost"), value, expected_status_code
        )

    def _query_agent(self, expected_value: int, agent_id: int) -> None:
        """
        Send a query to the ith agent about the state of the replicated counter.

        :param expected_value: the expected value of the counter.
        :param agent_id: the agent to be queried
        """
        node = self.tendermint_net_builder.nodes[agent_id]
        result = requests.get(
            node.get_http_addr("localhost") + "/abci_query",
            timeout=DEFAULT_REQUESTS_TIMEOUT,
        )
        assert result.status_code == 200
        counter_value_base64 = result.json()["result"]["response"]["value"].encode(
            "ascii"
        )
        counter_value_bytes = base64.b64decode(counter_value_base64)
        counter_value, *_ = struct.unpack(">I", counter_value_bytes)
        assert counter_value == expected_value


class TestABCICounterCrashFailureRestart(
    AEATestCaseMany, BaseTendermintTestClass, BaseTestABCICounterSkill
):
    """Test that restarting the agent with the same Tendermint node will restore the state."""

    NB_TX = 5

    def test_run(self) -> None:
        """Run the test."""
        agent_name = "counter_aea"
        self.fetch_agent("valory/counter:0.1.0", agent_name)
        self.set_agent_context(agent_name)
        self.generate_private_key("ethereum")
        self.add_private_key("ethereum", "ethereum_private_key.txt")
        self.set_config(
            "vendor.valory.connections.abci.config.tendermint_config.home",
            str(self.t / "tendermint_home"),
        )

        node_address = "http://localhost:26657"

        process = self.run_agent()
        is_running = self.is_running(process)
        assert is_running, "AEA not running within timeout!"

        time.sleep(5.0)

        for tx_number in range(self.NB_TX):
            time.sleep(2.0)
            new_value = tx_number + 1
            self._send_tx_to_node(node_address, new_value, 200)
        # wait synchronization
        time.sleep(5.0)

        # check that the final counter state is N
        self._query_node(node_address, self.NB_TX)

        check_strings = (*[f"The new count is: {x + 1}" for x in range(self.NB_TX)],)
        missing_strings = self.missing_from_output(process, check_strings)
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in agent output.".format(missing_strings)

        # Run it again
        process = self.run_agent()
        is_running = self.is_running(process)
        assert is_running, "AEA not running within timeout!"

        # give time to synchronize
        time.sleep(5.0)

        self._query_node(node_address, self.NB_TX)

        missing_strings = self.missing_from_output(process, check_strings)
        assert (
            missing_strings == []
        ), "Strings {} didn't appear in agent output.".format(missing_strings)

        if not self.is_successfully_terminated(process):
            warnings.warn(
                UserWarning(
                    f"ABCI agent with process {process} wasn't successfully terminated."
                )
            )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Tests for valory/registration_abci skill's behaviours."""

import time
from pathlib import Path
from typing import Any, Dict, Type, cast

from aea.helpers.transaction.base import (
    RawTransaction,
    SignedTransaction,
    TransactionDigest,
    TransactionReceipt,
)

from packages.open_aea.protocols.signing import SigningMessage
from packages.valory.contracts.offchain_aggregator.contract import (
    PUBLIC_ID as ORACLE_CONTRACT_ID,
)
from packages.valory.protocols.contract_api.message import ContractApiMessage
from packages.valory.protocols.ledger_api.message import LedgerApiMessage
from packages.valory.skills.abstract_round_abci.base import AbciAppDB
from packages.valory.skills.abstract_round_abci.behaviour_utils import (
    BaseBehaviour,
    make_degenerate_behaviour,
)
from packages.valory.skills.oracle_deployment_abci.behaviours import (
    DeployOracleBehaviour,
    RandomnessOracleBehaviour,
    SelectKeeperOracleBehaviour,
)
from packages.valory.skills.oracle_deployment_abci.behaviours import (
    SynchronizedData as OracleDeploymentSynchronizedSata,
)
from packages.valory.skills.oracle_deployment_abci.behaviours import (
    ValidateOracleBehaviour,
)
from packages.valory.skills.oracle_deployment_abci.rounds import (
    Event as OracleDeploymentEvent,
)
from packages.valory.skills.oracle_deployment_abci.rounds import FinishedOracleRound

from tests.conftest import ROOT_DIR
from tests.test_skills.base import FSMBehaviourBaseCase
from tests.test_skills.test_abstract_round_abci.test_common import (
    BaseRandomnessBehaviourTest,
    BaseSelectKeeperBehaviourTest,
)


class OracleDeploymentAbciBaseCase(FSMBehaviourBaseCase):
    """Base case for testing PriceEstimation FSMBehaviour."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "oracle_deployment_abci"
    )


class TestRandomnessOracle(BaseRandomnessBehaviourTest):
    """Test randomness safe."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "oracle_deployment_abci"
    )

    randomness_behaviour_class = RandomnessOracleBehaviour
    next_behaviour_class = SelectKeeperOracleBehaviour
    done_event = OracleDeploymentEvent.DONE


class TestSelectKeeperOracleBehaviour(BaseSelectKeeperBehaviourTest):
    """Test SelectKeeperBehaviour."""

    path_to_skill = Path(
        ROOT_DIR, "packages", "valory", "skills", "oracle_deployment_abci"
    )

    select_keeper_behaviour_class = SelectKeeperOracleBehaviour
    next_behaviour_class = DeployOracleBehaviour
    done_event = OracleDeploymentEvent.DONE


class BaseDeployBehaviourTest(FSMBehaviourBaseCase):
    """Base DeployBehaviourTest."""

    behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[BaseBehaviour]
    synchronized_data_kwargs: Dict
    contract_id: str
    done_event: Any

    def test_deployer_act(
        self,
    ) -> None:
        """Run tests."""
        participants = frozenset({self.skill.skill_context.agent_address, "a_1", "a_2"})
        most_voted_keeper_address = self.skill.skill_context.agent_address
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            OracleDeploymentSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            participants=participants,
                            most_voted_keeper_address=most_voted_keeper_address,
                            **self.synchronized_data_kwargs,
                        )
                    ),
                )
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()

        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_DEPLOY_TRANSACTION,
            ),
            contract_id=self.contract_id,
            response_kwargs=dict(
                performative=ContractApiMessage.Performative.RAW_TRANSACTION,
                callable="get_deploy_transaction",
                raw_transaction=RawTransaction(
                    ledger_id="ethereum",
                    body={"contract_address": "safe_contract_address"},
                ),
            ),
        )

        self.mock_signing_request(
            request_kwargs=dict(
                performative=SigningMessage.Performative.SIGN_TRANSACTION,
            ),
            response_kwargs=dict(
                performative=SigningMessage.Performative.SIGNED_TRANSACTION,
                signed_transaction=SignedTransaction(ledger_id="ethereum", body={}),
            ),
        )

        self.mock_ledger_api_request(
            request_kwargs=dict(
                performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,
            ),
            response_kwargs=dict(
                performative=LedgerApiMessage.Performative.TRANSACTION_DIGEST,
                transaction_digest=TransactionDigest(
                    ledger_id="ethereum", body="tx_hash"
                ),
            ),
        )

        self.mock_ledger_api_request(
            request_kwargs=dict(
                performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,
                transaction_digest=TransactionDigest(
                    ledger_id="ethereum", body="tx_hash"
                ),
            ),
            response_kwargs=dict(
                performative=LedgerApiMessage.Performative.TRANSACTION_RECEIPT,
                transaction_receipt=TransactionReceipt(
                    ledger_id="ethereum",
                    receipt={"contractAddress": "stub"},
                    transaction={},
                ),
            ),
        )

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(self.done_event)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id

    def test_not_deployer_act(
        self,
    ) -> None:
        """Run tests."""
        participants = frozenset({self.skill.skill_context.agent_address, "a_1", "a_2"})
        most_voted_keeper_address = "a_1"
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            OracleDeploymentSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(
                        dict(
                            participants=participants,
                            most_voted_keeper_address=most_voted_keeper_address,
                            **self.synchronized_data_kwargs,
                        )
                    ),
                )
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()
        self._test_done_flag_set()
        self.end_round(self.done_event)
        time.sleep(1)
        self.behaviour.act_wrapper()
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestDeployOracleBehaviour(BaseDeployBehaviourTest, OracleDeploymentAbciBaseCase):
    """Test DeployOracleBehaviour."""

    behaviour_class = DeployOracleBehaviour
    next_behaviour_class = ValidateOracleBehaviour
    synchronized_data_kwargs = dict(
        safe_contract_address="safe_contract_address",
        oracle_contract_address="oracle_contract_address",
    )
    contract_id = str(ORACLE_CONTRACT_ID)
    done_event = OracleDeploymentEvent.DONE


class BaseValidateBehaviourTest(FSMBehaviourBaseCase):
    """Test ValidateSafeBehaviour."""

    behaviour_class: Type[BaseBehaviour]
    next_behaviour_class: Type[BaseBehaviour]
    synchronized_data_kwargs: Dict
    contract_id: str
    done_event: Any

    def test_validate_behaviour(self) -> None:
        """Run test."""
        self.fast_forward_to_behaviour(
            self.behaviour,
            self.behaviour_class.behaviour_id,
            OracleDeploymentSynchronizedSata(
                AbciAppDB(
                    setup_data=AbciAppDB.data_to_lists(self.synchronized_data_kwargs)
                ),
            ),
        )
        assert (
            cast(
                BaseBehaviour,
                cast(BaseBehaviour, self.behaviour.current_behaviour),
            ).behaviour_id
            == self.behaviour_class.behaviour_id
        )
        self.behaviour.act_wrapper()
        self.mock_contract_api_request(
            request_kwargs=dict(
                performative=ContractApiMessage.Performative.GET_STATE,
            ),
            contract_id=self.contract_id,
            response_kwargs=dict(
                performative=ContractApiMessage.Performative.STATE,
                callable="verify_contract",
                state=ContractApiMessage.State(
                    ledger_id="ethereum", body={"verified": True}
                ),
            ),
        )

        self.mock_a2a_transaction()
        self._test_done_flag_set()
        self.end_round(self.done_event)
        behaviour = cast(BaseBehaviour, self.behaviour.current_behaviour)
        assert behaviour.behaviour_id == self.next_behaviour_class.behaviour_id


class TestValidateOracleBehaviour(
    BaseValidateBehaviourTest, OracleDeploymentAbciBaseCase
):
    """Test ValidateOracleBehaviour."""

    behaviour_class = ValidateOracleBehaviour
    next_behaviour_class = make_degenerate_behaviour(FinishedOracleRound.round_id)
    synchronized_data_kwargs = dict(
        safe_contract_address="safe_contract_address",
        oracle_contract_address="oracle_contract_address",
    )
    contract_id = str(ORACLE_CONTRACT_ID)
    done_event = OracleDeploymentEvent.DONE



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests of the ledger API connection module."""
import asyncio
import logging
from typing import Any, Dict, Optional, cast
from unittest.mock import Mock, patch

import pytest
from aea.common import Address
from aea.configurations.data_types import PublicId
from aea.connections.base import Connection, ConnectionStates
from aea.crypto.ledger_apis import LedgerApis
from aea.crypto.registries import make_crypto, make_ledger_api
from aea.helpers.async_utils import AsyncState
from aea.helpers.transaction.base import (
    RawTransaction,
    SignedTransaction,
    Terms,
    TransactionDigest,
    TransactionReceipt,
)
from aea.mail.base import Envelope, Message
from aea.protocols.dialogue.base import Dialogue as BaseDialogue
from aea_ledger_ethereum import EthereumCrypto

from packages.valory.connections.ledger.connection import LedgerConnection
from packages.valory.connections.ledger.ledger_dispatcher import (
    LedgerApiRequestDispatcher,
)
from packages.valory.protocols.ledger_api.custom_types import Kwargs
from packages.valory.protocols.ledger_api.dialogues import LedgerApiDialogue
from packages.valory.protocols.ledger_api.dialogues import (
    LedgerApiDialogues as BaseLedgerApiDialogues,
)
from packages.valory.protocols.ledger_api.message import LedgerApiMessage

from tests.conftest import ETHEREUM_KEY_DEPLOYER
from tests.helpers.docker.ganache import DEFAULT_GANACHE_CHAIN_ID


SOME_SKILL_ID = "some/skill:0.1.0"

logger = logging.getLogger(__name__)
ledger_ids = pytest.mark.parametrize(
    "ledger_id,address",
    [
        (EthereumCrypto.identifier, EthereumCrypto(ETHEREUM_KEY_DEPLOYER).address),
    ],
)
gas_strategies = pytest.mark.parametrize(
    "gas_strategies",
    [
        {"gas_price_strategy": None},
        {"gas_price_strategy": "gas_station"},
        {"gas_price_strategy": "eip1559"},
        {
            "max_fee_per_gas": 1_000_000_000,
            "max_priority_fee_per_gas": 1_000_000_000,
        },
    ],
)


class LedgerApiDialogues(BaseLedgerApiDialogues):
    """The dialogues class keeps track of all ledger_api dialogues."""

    def __init__(self, self_address: Address, **kwargs: Any) -> None:
        """Initialize dialogues."""

        def role_from_first_message(  # pylint: disable=unused-argument
            message: Message, receiver_address: Address
        ) -> BaseDialogue.Role:
            """Infer the role of the agent from an incoming/outgoing first message"""
            return LedgerApiDialogue.Role.AGENT

        BaseLedgerApiDialogues.__init__(
            self,
            self_address=self_address,
            role_from_first_message=role_from_first_message,
        )


@pytest.mark.usefixtures("ganache_scope_class")
class TestLedgerConnection:
    """Tests for ledger connection."""

    @pytest.mark.asyncio
    @ledger_ids
    async def test_get_balance(
        self,
        ledger_id: str,
        address: str,
        ledger_apis_connection: Connection,
        update_default_ethereum_ledger_api: None,
        ethereum_testnet_config: Dict,
    ) -> None:
        """Test get balance method."""

        config = ethereum_testnet_config
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_BALANCE,  # type: ignore
            ledger_id=ledger_id,
            address=address,
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_msg = cast(LedgerApiMessage, response.message)
        response_dialogue = ledger_api_dialogues.update(response_msg)
        assert response_dialogue == ledger_api_dialogue
        assert response_msg.performative == LedgerApiMessage.Performative.BALANCE
        actual_balance_amount = response_msg.balance
        expected_balance_amount = make_ledger_api(ledger_id, **config).get_balance(
            address
        )
        assert actual_balance_amount == expected_balance_amount

    @pytest.mark.asyncio
    @ledger_ids
    async def test_get_state(
        self,
        ledger_id: str,
        address: str,
        ledger_apis_connection: Connection,
        update_default_ethereum_ledger_api: None,
        ethereum_testnet_config: Dict,
    ) -> None:
        """Test get state."""

        config = ethereum_testnet_config

        if "ethereum" in ledger_id:
            callable_name = "get_block"
        else:
            callable_name = "blocks"
        args = ("latest",)
        kwargs = Kwargs({})

        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_STATE,  # type: ignore
            ledger_id=ledger_id,
            callable=callable_name,
            args=args,
            kwargs=kwargs,
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_msg = cast(LedgerApiMessage, response.message)
        response_dialogue = ledger_api_dialogues.update(response_msg)
        assert response_dialogue == ledger_api_dialogue

        assert (
            response_msg.performative == LedgerApiMessage.Performative.STATE
        ), response_msg
        actual_block = response_msg.state.body
        expected_block = make_ledger_api(ledger_id, **config).get_state(
            callable_name, *args
        )
        assert actual_block == expected_block

    @pytest.mark.asyncio
    @gas_strategies
    async def test_get_raw_transaction(
        self,
        gas_strategies: Dict,
        ledger_apis_connection: Connection,
        update_default_ethereum_ledger_api: None,
    ) -> None:
        """Test get raw transaction with Ethereum APIs."""
        import aea  # noqa # to load registries

        crypto1 = make_crypto(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_DEPLOYER
        )
        crypto2 = make_crypto(EthereumCrypto.identifier)
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)

        amount = 40000
        fee = 10 ** 7

        # Create ledger_api dialogue: get raw transaction
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            terms=Terms(
                ledger_id=EthereumCrypto.identifier,
                sender_address=crypto1.address,
                counterparty_address=crypto2.address,
                amount_by_currency_id={"ETH": -amount},
                quantities_by_good_id={"some_service_id": 1},
                is_sender_payable_tx_fee=True,
                nonce="",
                fee_by_currency_id={"ETH": fee},
                chain_id=DEFAULT_GANACHE_CHAIN_ID,
                **gas_strategies,
            ),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )
        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)

        # Raw transaction
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.RAW_TRANSACTION
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.raw_transaction, RawTransaction)
        assert response_message.raw_transaction.ledger_id == request.terms.ledger_id

    @pytest.mark.asyncio
    @gas_strategies
    async def test_send_signed_transaction_ethereum(
        self, gas_strategies: Dict, ledger_apis_connection: LedgerConnection
    ) -> None:
        """Test send signed transaction with Ethereum APIs."""
        ledger_api_dialogues = LedgerApiDialogues(SOME_SKILL_ID)

        crypto1 = make_crypto(
            EthereumCrypto.identifier, private_key_path=ETHEREUM_KEY_DEPLOYER
        )
        crypto2 = make_crypto(EthereumCrypto.identifier)

        # First, send a transaction so we can get a digest at the end
        amount = 40000
        fee = 10 ** 7

        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            terms=Terms(
                ledger_id=EthereumCrypto.identifier,
                sender_address=crypto1.address,
                counterparty_address=crypto2.address,
                amount_by_currency_id={"ETH": -amount},
                quantities_by_good_id={"some_service_id": 1},
                is_sender_payable_tx_fee=True,
                nonce="",
                fee_by_currency_id={"ETH": fee},
                chain_id=DEFAULT_GANACHE_CHAIN_ID,
                **gas_strategies,
            ),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        # Check that we got the correct transaction response
        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.RAW_TRANSACTION
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.raw_transaction, RawTransaction)
        assert response_message.raw_transaction.ledger_id == request.terms.ledger_id

        # Sign the transaction
        signed_transaction = crypto1.sign_transaction(
            response_message.raw_transaction.body
        )

        # Create new dialogue starting with signed transaction
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.SEND_SIGNED_TRANSACTION,  # type: ignore
            signed_transaction=SignedTransaction(
                EthereumCrypto.identifier, signed_transaction
            ),
        )
        request = cast(LedgerApiMessage, request)
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )

        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)

        # Transaction digest
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative != LedgerApiMessage.Performative.ERROR
        ), f"Received error: {response_message.message}"
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.TRANSACTION_DIGEST
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.transaction_digest, TransactionDigest)
        assert isinstance(response_message.transaction_digest.body, str)
        assert (
            response_message.transaction_digest.ledger_id
            == request.signed_transaction.ledger_id
        )

        # Create new dialogue starting with GET_TRANSACTION_RECEIPT
        request, ledger_api_dialogue = ledger_api_dialogues.create(
            counterparty=str(ledger_apis_connection.connection_id),
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,  # type: ignore
            transaction_digest=response_message.transaction_digest,
        )
        envelope = Envelope(
            to=request.to,
            sender=request.sender,
            message=request,
        )
        await ledger_apis_connection.send(envelope)
        await asyncio.sleep(0.01)

        # Transaction receipt
        response = await ledger_apis_connection.receive()

        assert response is not None
        assert isinstance(response.message, LedgerApiMessage)
        response_message = cast(LedgerApiMessage, response.message)
        assert (
            response_message.performative
            == LedgerApiMessage.Performative.TRANSACTION_RECEIPT
        )
        response_dialogue = ledger_api_dialogues.update(response_message)
        assert response_dialogue == ledger_api_dialogue
        assert isinstance(response_message.transaction_receipt, TransactionReceipt)
        assert response_message.transaction_receipt.receipt is not None
        assert response_message.transaction_receipt.transaction is not None
        assert (
            response_message.transaction_receipt.ledger_id
            == request.transaction_digest.ledger_id  # type: ignore
        )
        assert LedgerApis.is_transaction_settled(
            response_message.transaction_receipt.ledger_id,
            response_message.transaction_receipt.receipt,
        ), "Transaction not settled."

    @pytest.mark.asyncio
    async def test_unsupported_protocol(
        self, ledger_apis_connection: LedgerConnection
    ) -> None:
        """Test fail on protocol not supported."""
        envelope = Envelope(
            to=str(ledger_apis_connection.connection_id),
            sender="test/skill:0.1.0",
            protocol_specification_id=PublicId.from_str("author/package_name:0.1.0"),
            message=b"message",
        )
        with pytest.raises(ValueError, match="Protocol not supported"):
            ledger_apis_connection._schedule_request(envelope)

    @pytest.mark.asyncio
    async def test_no_balance(
        self,
    ) -> None:
        """Test no balance."""
        dispatcher = LedgerApiRequestDispatcher(AsyncState())
        mock_api = Mock()
        message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.GET_BALANCE,  # type: ignore
            dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
            ledger_id=EthereumCrypto.identifier,
            address="test",
        )
        message.to = dispatcher.dialogues.self_address
        message.sender = "test"
        dialogue = cast(
            Optional[LedgerApiDialogue], dispatcher.dialogues.update(message)
        )
        assert dialogue is not None

        mock_api.get_balance.return_value = None
        msg = dispatcher.get_balance(mock_api, message, dialogue)
        assert msg.performative == LedgerApiMessage.Performative.ERROR

    @pytest.mark.asyncio
    async def test_no_raw_tx(
        self,
    ) -> None:
        """Test no raw tx returned."""
        dispatcher = LedgerApiRequestDispatcher(AsyncState())
        mock_api = Mock()
        message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.GET_RAW_TRANSACTION,  # type: ignore
            dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
            terms=Terms(
                ledger_id=EthereumCrypto.identifier,
                sender_address="1111",
                counterparty_address="22222",
                amount_by_currency_id={"ETH": -1},
                quantities_by_good_id={"some_service_id": 1},
                is_sender_payable_tx_fee=True,
                nonce="",
                fee_by_currency_id={"ETH": 10},
                chain_id=3,
            ),
        )
        message.to = dispatcher.dialogues.self_address
        message.sender = "test"
        dialogue = cast(
            Optional[LedgerApiDialogue], dispatcher.dialogues.update(message)
        )
        assert dialogue is not None

        mock_api.get_transfer_transaction.return_value = None
        msg = dispatcher.get_raw_transaction(mock_api, message, dialogue)
        assert msg.performative == LedgerApiMessage.Performative.ERROR

    @pytest.mark.asyncio
    async def test_attempts_get_transaction_receipt(
        self,
    ) -> None:
        """Test retry and sleep."""
        dispatcher = LedgerApiRequestDispatcher(AsyncState(ConnectionStates.connected))
        mock_api = Mock()
        message = LedgerApiMessage(
            performative=LedgerApiMessage.Performative.GET_TRANSACTION_RECEIPT,  # type: ignore
            dialogue_reference=dispatcher.dialogues.new_self_initiated_dialogue_reference(),
            transaction_digest=TransactionDigest("asdad", "sdfdsf"),
        )
        message.to = dispatcher.dialogues.self_address
        message.sender = "test"
        dialogue = cast(
            Optional[LedgerApiDialogue], dispatcher.dialogues.update(message)
        )
        assert dialogue is not None
        mock_api.get_transaction.return_value = None
        mock_api.is_transaction_settled.return_value = True
        with patch.object(dispatcher, "retry_attempts", 2):
            with patch.object(dispatcher, "retry_timeout", 0.001):
                msg = dispatcher.get_transaction_receipt(mock_api, message, dialogue)

        assert msg.performative == LedgerApiMessage.Performative.ERROR



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""This module contains the tests for the code-blocks in the documentation."""

from pathlib import Path
from typing import Dict, List, Optional

from tests.conftest import ROOT_DIR
from tests.test_docs.helper import (  # type: ignore
    CodeType,
    check_bash_commands_exist,
    check_code_blocks_exist,
    contains_code_blocks,
    extract_autonomy_commands,
    extract_make_commands,
    remove_doc_ellipsis,
    remove_ips_hashes,
    remove_line_comments,
)


class BaseTestDocCode:
    """Base class for doc code testing"""

    md_to_code: Dict[str, Dict] = {}
    code_type: CodeType = CodeType.NOCODE
    skipped_files: Optional[List[str]] = None

    def test_run_check(self) -> None:
        """Check the documentaion code"""

        assert (
            self.code_type != CodeType.NOCODE
        ), "This test class has not specified a code type to check for"
        assert (
            self.md_to_code != {}
        ), "This test class has not specified the md_to_code mapping"

        # Get all doc files that contain a block
        all_md_files = [
            str(p.relative_to(ROOT_DIR)) for p in Path(ROOT_DIR, "docs").rglob("*.md")
        ]
        files_with_blocks = list(
            filter(
                lambda f: "/api/" not in f  # skip api folder
                and contains_code_blocks(f, self.code_type.value),
                all_md_files,
            )
        )
        if self.skipped_files:
            files_with_blocks = [
                f for f in files_with_blocks if f not in self.skipped_files
            ]
        not_checked_files = set(files_with_blocks).difference(
            set(self.md_to_code.keys())
        )

        assert (
            not not_checked_files
        ), f"The following (not skipped) doc files contain {self.code_type.value} blocks but are not being checked: {not_checked_files}"

        # Check all files
        for md_file, code_info in self.md_to_code.items():
            print(
                f"Checking {self.code_type.value} snippets in file {md_file}... ",
                end="",
            )

            # In doc files: remove tokens like "# ...\n" from the code
            # In code files: replace ipfs hashes with a placeholder
            check_code_blocks_exist(
                md_file=md_file,
                code_info=code_info,
                code_type=self.code_type,
                doc_process_fn=lambda s: remove_doc_ellipsis(remove_line_comments(s)),
                code_process_fn=lambda s: remove_ips_hashes(s),
            ),

            print("OK")


class TestYamlSnippets(BaseTestDocCode):
    """Test that all the yaml snippets in the documentation exist in the repository"""

    code_type = CodeType.YAML

    # This variable holds a mapping between every doc file and the code file
    # that contains the referenced code. Since a doc file can contain several code
    # snippets, a list with the target files ordered is provided.
    #
    # Use skip_blocks to specify a list of blocks that need to be skipped
    # Add by_line:: at the beggining of a code file path so the check is performed line by line
    # instead of checking the code block as a whole.

    md_to_code = {
        "docs/get_started.md": {
            "code_files": ["packages/valory/agents/hello_world/aea-config.yaml"],
        },
        "docs/price_oracle_fsms.md": {
            "code_files": [
                "packages/valory/skills/registration_abci/fsm_specification.yaml",
                "packages/valory/skills/safe_deployment_abci/fsm_specification.yaml",
                "packages/valory/skills/oracle_deployment_abci/fsm_specification.yaml",
                "packages/valory/skills/price_estimation_abci/fsm_specification.yaml",
                "packages/valory/skills/transaction_settlement_abci/fsm_specification.yaml",
                "packages/valory/skills/reset_pause_abci/fsm_specification.yaml",
                "packages/valory/skills/oracle_abci/fsm_specification.yaml",
            ],
        },
        "docs/simple_abci.md": {
            "code_files": ["packages/valory/skills/simple_abci/fsm_specification.yaml"],
        },
        "docs/networks.md": {"skip_blocks": [0]},
    }


class TestPythonSnippets(BaseTestDocCode):
    """Test that all the python snippets in the documentation exist in the repository"""

    code_type = CodeType.PYTHON

    # This variable holds a mapping between every doc file and the code file
    # that contains the referenced code. Since a doc file can contain several code
    # snippets, a list with the target files ordered is provided.
    #
    # Use skip_blocks to specify a list of blocks that need to be skipped
    # Add by_line:: at the beggining of a code file path so the check is performed line by line
    # instead of checking the code block as a whole.

    md_to_code = {
        "docs/abci_app_abstract_round_behaviour.md": {
            "code_files": [
                "by_line::packages/valory/skills/abstract_round_abci/behaviours.py"
            ],
            "skip_blocks": [1],
        },
        "docs/abci_app_class.md": {
            "code_files": [
                "by_line::packages/valory/skills/abstract_round_abci/base.py"
            ],
            "skip_blocks": [1],
        },
        "docs/get_started.md": {
            "code_files": [
                "by_line::packages/valory/skills/hello_world_abci/rounds.py",
                "by_line::packages/valory/skills/hello_world_abci/rounds.py",
                "packages/valory/skills/hello_world_abci/behaviours.py",
                "packages/valory/skills/hello_world_abci/behaviours.py",
                "packages/valory/skills/hello_world_abci/payloads.py",
            ],
        },
        "docs/price_oracle_intro.md": {
            "code_files": [
                "packages/valory/skills/oracle_abci/composition.py",
                "packages/valory/skills/oracle_abci/behaviours.py",
            ],
        },
        "docs/price_oracle_fsms.md": {
            "code_files": [
                "packages/valory/skills/oracle_abci/composition.py",
                "packages/valory/skills/oracle_abci/composition.py",
            ],
        },
        "docs/simple_abci.md": {
            "code_files": [
                "packages/valory/skills/simple_abci/behaviours.py",
                "packages/valory/skills/simple_abci/rounds.py",
            ],
        },
    }

    skipped_files = [
        "docs/abci_app_async_behaviour.md",  # just placeholder examples
        "docs/networks.md",  # only irrelevant one-liners
    ]


class TestDocBashSnippets:
    """Class for doc bash snippet testing"""

    def test_run_check(self) -> None:
        """Check the documentaion code"""

        code_type = CodeType.BASH

        skipped_files: List[str] = []

        # Get all doc files that contain a block
        all_md_files = [
            str(p.relative_to(ROOT_DIR)) for p in Path(ROOT_DIR, "docs").rglob("*.md")
        ]
        files_with_blocks = list(
            filter(
                lambda f: "/api/" not in f  # skip api folder
                and contains_code_blocks(f, code_type.value),
                all_md_files,
            )
        )
        if skipped_files:
            files_with_blocks = [f for f in files_with_blocks if f not in skipped_files]

        all_mk_files = [
            str(p.relative_to(ROOT_DIR)) for p in Path(ROOT_DIR).rglob("*Makefile")
        ]

        all_mk_files = list(
            filter(
                lambda f: "third_party/" not in f
                and ".tox" not in f,  # skip some folders
                all_mk_files,
            )
        )

        make_commands = extract_make_commands(all_mk_files)
        autonomy_commands = extract_autonomy_commands()

        for md_file in files_with_blocks:
            check_bash_commands_exist(md_file, make_commands, autonomy_commands)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Analyse ABCI app definitions for docstrings."""

import importlib
import re
from pathlib import Path
from typing import Any, cast


INDENT = " " * 4
NEWLINE = "\n"
COMMA = ", "
ABCIAPP = "AbciApp"
DOCSTRING_TEMPLATE = """
\"\"\"{abci_app_name}\n
Initial round: {initial_round}\n
Initial states: {{{initial_states}}}\n
Transition states:
    {transition_states}\n
Final states: {{{final_states}}}\n
Timeouts:
{timeouts}
\"\"\""""


def docstring_abci_app(abci_app: Any) -> str:  # pylint: disable-msg=too-many-locals
    """
    Generate a docstring for an ABCI app

    This ensures that documentation aligns with the actual implementation

    :param abci_app: abci app object.
    :return: docstring
    """

    states = {state: i for i, state in enumerate(abci_app.transition_function)}
    initial_round = abci_app.initial_round_cls.__name__
    initial_states = [state.__name__ for state in abci_app.initial_states]
    initial_states = initial_states if initial_states else [initial_round]
    final_states = [state.__name__ for state in abci_app.final_states]

    transition_states = []
    for state, transitions in abci_app.transition_function.items():
        transition_states.append(f"{states[state]}. {state.__name__}")
        for event, next_state in transitions.items():
            name = event.value.replace("_", " ")
            transition_states.append(f"{INDENT}- {name}: {states[next_state]}.")

    timeouts = []
    for event, seconds in abci_app.event_to_timeout.items():
        timeouts.append(f"{INDENT}{event.value.replace('_', ' ')}: {seconds}")

    return DOCSTRING_TEMPLATE.format(
        abci_app_name=abci_app.__name__,  # type: ignore
        initial_round=initial_round,
        initial_states=COMMA.join(sorted(initial_states)),
        transition_states=(NEWLINE + INDENT).join(transition_states),
        final_states=COMMA.join(sorted(final_states)),
        timeouts=NEWLINE.join(timeouts),
    )


def update_docstrings(
    module_path: Path, docstring: str, abci_app_name: str, check: bool = False
) -> bool:
    """Update docstrings."""

    content = module_path.read_text()
    regex = r'class [A-Za-z]+\(AbciApp\[Event\]\):([a-zA-Z \#:=\-]+)?\n    """[A-Za-z]+\n[a-zA-Z0-9 :{},._\-\n]+"""'
    docstring = "\n".join(
        map(lambda x: f"{INDENT}{x}" if len(x) else x, docstring.split("\n"))
    )
    match = re.search(regex, content)
    if match is not None:
        group, *_ = cast(re.Match, match).groups()
        markers = group if group is not None else ""

    updated_class = f"class {abci_app_name}(AbciApp[Event]):{markers}{docstring}"
    updated_content = re.sub(regex, updated_class, content, re.MULTILINE)
    needs_update = updated_content != content

    if check:
        return needs_update

    module_path.write_text(updated_content, encoding="utf-8")
    return needs_update


def process_module(module_path: Path, check: bool = False) -> bool:
    """Process module."""
    module = importlib.import_module(".".join(module_path.parts).replace(".py", ""))
    for obj in dir(module):
        if obj.endswith(ABCIAPP) and obj != ABCIAPP:
            App = getattr(module, obj)
            docstring = docstring_abci_app(App)
            return update_docstrings(module_path, docstring, obj, check)

    return False



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Script for generating deployment environments."""
from pathlib import Path
from typing import Optional, cast

from autonomy.constants import (
    HARDHAT_IMAGE_VERSION,
    IMAGE_VERSION,
    TENDERMINT_IMAGE_VERSION,
)
from autonomy.deploy.base import BaseDeploymentGenerator, ServiceSpecification
from autonomy.deploy.constants import DEPLOYMENT_REPORT
from autonomy.deploy.generators.docker_compose.base import DockerComposeGenerator
from autonomy.deploy.generators.kubernetes.base import KubernetesGenerator


DEPLOYMENT_OPTIONS = {
    "kubernetes": KubernetesGenerator,
    "docker-compose": DockerComposeGenerator,
}


def generate_deployment(  # pylint: disable=too-many-arguments
    type_of_deployment: str,
    private_keys_file_path: Path,
    service_path: Path,
    packages_dir: Path,
    build_dir: Path,
    number_of_agents: Optional[int] = None,
    private_keys_password: Optional[str] = None,
    dev_mode: bool = False,
    version: Optional[str] = None,
) -> str:
    """Generate the deployment build for the valory app."""

    if version is None:
        image_versions = {
            "agent": IMAGE_VERSION,
            "hardhat": HARDHAT_IMAGE_VERSION,
            "tendermint": TENDERMINT_IMAGE_VERSION,
        }
    else:
        image_versions = {
            "agent": version,
            "hardhat": version,
            "tendermint": version,
        }

    service_spec = ServiceSpecification(
        service_path=service_path,
        keys=private_keys_file_path,
        packages_dir=packages_dir,
        private_keys_password=private_keys_password,
        number_of_agents=number_of_agents,
    )

    DeploymentGenerator = DEPLOYMENT_OPTIONS.get(type_of_deployment)
    if DeploymentGenerator is None:
        raise ValueError(f"Cannot find deployment generator for {type_of_deployment}")
    deployment = cast(
        BaseDeploymentGenerator,
        DeploymentGenerator(service_spec=service_spec, build_dir=build_dir),
    )

    deployment.generate(image_versions, dev_mode).generate_config_tendermint(
        image_versions["tendermint"]
    ).write_config().populate_private_keys()

    return DEPLOYMENT_REPORT.substitute(
        **{
            "type": type_of_deployment,
            "agents": service_spec.service.number_of_agents,
            "network": service_spec.service.network,
            "size": len(deployment.output),
        }
    )



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Image building."""

import os
import shutil
import signal
import subprocess  # nosec
from pathlib import Path
from typing import IO, cast

import yaml
from aea.configurations.utils import PublicId


class ImageProfiles:  # pylint: disable=too-few-public-methods
    """Image build profiles."""

    CLUSTER = "cluster"
    DEVELOPMENT = "dev"
    PRODUCTION = "prod"
    DEPENDENCIES = "dependencies"
    ALL = (CLUSTER, DEVELOPMENT, PRODUCTION, DEPENDENCIES)


class ImageBuilder:
    """Class to build images using skaffold."""

    @classmethod
    def build_images(  # pylint: disable=too-many-arguments
        cls,
        profile: str,
        deployment_file_path: Path,
        packages_dir: Path,
        build_dir: Path,
        skaffold_dir: Path,
        version: str,
        push: bool = False,
    ) -> None:
        """Build images using the subprocess."""

        if profile == ImageProfiles.DEVELOPMENT:
            version = "dev"

        aea_agent = cls.get_aea_agent(deployment_file_path=deployment_file_path)
        cls._copy_packages(packages_dir=packages_dir, build_dir=build_dir)
        cls._build(aea_agent, profile, skaffold_dir, version, push)

    @classmethod
    def _build(  # pylint: disable=too-many-arguments
        cls,
        aea_agent: str,
        profile: str,
        skaffold_dir: Path,
        version: str,
        push: bool = False,
        build_concurrency: int = 0,
    ) -> None:
        """Command to build images from for skaffold deployment."""

        agent_id = PublicId.from_str(aea_agent)
        env = os.environ.copy()
        env["AEA_AGENT"] = aea_agent

        if profile == ImageProfiles.DEPENDENCIES:
            env["VERSION"] = version
        else:
            env["VERSION"] = f"{agent_id.name}-{version}"

        if profile == ImageProfiles.CLUSTER:
            if env.get("KUBECONFIG") is None:
                raise ValueError("Please setup KUBECONFIG variable.")
        else:
            # deleting KUBECONFIG var for none cluster builds will avoid uneccessary warnings
            if env.get("KUBECONFIG") is not None:
                del env["KUBECONFIG"]

        try:
            process = (
                subprocess.Popen(  # nosec # pylint: disable=consider-using-with,W1509
                    [
                        "skaffold",
                        "build",
                        f"--build-concurrency={build_concurrency}",
                        f"--push={str(push).lower()}",
                        "-p",
                        profile,
                    ],
                    preexec_fn=os.setsid,
                    env=env,
                    stdout=subprocess.PIPE,
                    cwd=str(skaffold_dir),
                )
            )

            for line in iter(cast(IO[bytes], process.stdout).readline, ""):
                if line == b"":
                    break
                print(f"[Skaffold] {line.decode().strip()}")

        except KeyboardInterrupt:
            cast(IO[bytes], process.stdout).close()
            process.send_signal(signal.SIGTERM)

        process.wait(timeout=30)
        poll = process.poll()
        if poll is None:
            process.terminate()
            process.wait(2)

        if process.returncode != 0:
            print("Image build failed.")

    @staticmethod
    def _copy_packages(
        packages_dir: Path,
        build_dir: Path,
    ) -> None:
        """Copy packages for image building."""
        build_packages_dir = Path(build_dir) / "packages"

        if build_packages_dir.exists():
            shutil.rmtree(build_packages_dir)

        shutil.copytree(  # type: ignore # pylint: disable=E1123
            src=Path(packages_dir), dst=build_packages_dir
        )

    @staticmethod
    def get_aea_agent(deployment_file_path: Path) -> str:
        """Validate and retrieve aea agent from spec."""

        with open(deployment_file_path, "r", encoding="utf-8") as stream:
            deployment_spec, *_ = yaml.safe_load_all(stream)
            agent = deployment_spec["agent"]

        return agent



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Core for cli."""

import click
from aea.cli.core import cli as aea_cli

from autonomy.cli.analyse import analyse_group
from autonomy.cli.deploy import deploy_group
from autonomy.cli.develop import develop_group
from autonomy.cli.hash import hash_group
from autonomy.cli.replay import replay_group


@click.group(name="autonomy")  # type: ignore
@click.pass_context
def autonomy_cli(
    click_context: click.Context,  # pylint: disable=unused-argument
) -> None:
    """Command-line tool for setting up an autonomys of AEAs."""


autonomy_cli.add_command(analyse_group)
autonomy_cli.add_command(deploy_group)
autonomy_cli.add_command(develop_group)
autonomy_cli.add_command(replay_group)
aea_cli.add_command(hash_group)

cli = click.CommandCollection(sources=[aea_cli, autonomy_cli])



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""HTML templates for benchmark outputs."""

HTML_TEMPLATE = """<html>
<head>
    <title>
        Benchmarks
    </title>
    <style>
        *{{
            padding: 0;
            margin: 0;
        }}
        body{{
            padding: 5;
        }}
        td{{
            height: 24px;
            min-width: 96px;
            max-width: fit-content;
            font-size: 14px;
            padding: 3px;
        }}

        th{{
            height: 24px;
            min-width: 96px;
            max-width: fit-content;
            font-size: 14px;
            padding: 3px;
        }}
    </style>
</head>
<body>
    {tables}
</body>
</hmtl>"""

TABLE_TEMPLATE = """
<div style="margin: 16px 8px;">
    <div style="height: 3vh; width: 100%; font-size: 14px;">
        Period: {period} | Block: {block}
    </div>
    {table}
</div>"""



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""Test the serializer.py module of the skill."""

import math
import sys
from collections import defaultdict
from typing import Any, Dict

import atheris  # type: ignore
import hypothesis.strategies as st
import pytest
from google.protobuf.struct_pb2 import Struct
from hypothesis import given

from packages.valory.skills.abstract_round_abci import serializer
from packages.valory.skills.abstract_round_abci.serializer import (
    DictProtobufStructSerializer,
)


def test_encode_decode_i() -> None:
    """Test encode decode logic."""
    case = {
        "key1": True,
        "key2": 0.12,
        "key3": 100,
        "key4": "some string",
        "key5": b"some bytes string",
        "key6": Struct(),
        "_need_patch": {},
    }
    encoded = DictProtobufStructSerializer.encode(case)
    assert isinstance(encoded, bytes)
    decoded = DictProtobufStructSerializer.decode(encoded)
    assert case == decoded


def test_encode_decode_ii() -> None:
    """Test encode decode logic."""
    case = {
        "key1": True,
        "key2": 0.12,
        "key3": 100,
        "key4": "some string",
        "key5": b"some bytes string",
        "key6": {"key1": True, "key2": 0.12},
    }
    encoded = DictProtobufStructSerializer.encode(case)
    assert isinstance(encoded, bytes)
    decoded = DictProtobufStructSerializer.decode(encoded)
    assert case == decoded


# utility functions
def node() -> defaultdict:
    """Recursive defaultdict"""
    return defaultdict(node)


def to_dict(dd: Dict[str, Any]) -> Dict[str, Any]:
    """Recursive defaultdict to dict"""
    return {k: to_dict(v) for k, v in dd.items()} if isinstance(dd, defaultdict) else dd


def types_of(d: Dict[str, Any]) -> Dict[str, Any]:
    """Get `key: type(value)` mapping, recursively."""
    return {k: types_of(v) if isinstance(v, dict) else type(v) for k, v in d.items()}


def is_decodable(b: bytes) -> bool:
    """Check if bytes can be decoded"""
    try:
        b.decode(serializer.ENCODING)
        return True
    except UnicodeDecodeError:
        return False


def is_serializer_compatible(data: Dict) -> bool:
    """Check whether the serializer can reconstitute the data"""
    serialized = serializer.to_bytes(data)
    assert isinstance(serialized, bytes)
    deserialized = serializer.from_bytes(serialized)
    return data == deserialized and types_of(data) == types_of(deserialized)


# tests
@pytest.mark.parametrize(
    "unsupported_type",
    [bool, int, float, tuple, frozenset],
)
def test_unsupported_key_types(unsupported_type: Any) -> None:
    """Python accepted key-types not compatible with protobuf"""
    data = {unsupported_type(): "value"}
    with pytest.raises(Exception):
        serializer.to_bytes(data)


@pytest.mark.parametrize(
    "unsupported_type",
    [tuple, list, set, frozenset],
)
def test_unsupported_value_type(unsupported_type: Any) -> None:
    """Not implemented."""
    data = {"key": unsupported_type()}
    with pytest.raises(NotImplementedError):
        serializer.to_bytes(data)


@pytest.mark.parametrize(
    "value",
    [True, 1 << 256, 3.14, "string", b"bytes", {}],
)
def test_single_values(value: Any) -> None:
    """Single value type test"""
    assert is_serializer_compatible({"key": value})


def test_nested_mapping() -> None:
    """Nested mapping test"""
    root = node()
    root["a"]["0"] = {}
    root["a"]["1"] = False
    root["a"]["2"] = -(1 << 256)
    root["a"]["3"]["i"] = -3.141592653589793
    root["a"]["3"]["ii"] = b"bytes"
    root["a"]["3"]["iii"] = "string"
    assert is_serializer_compatible(to_dict(root))


# randomized hypothesis testing
value_strategy = st.one_of(
    st.booleans(),
    st.integers(),
    st.floats(allow_nan=False),  # because `nan != nan`
    st.text(),
    st.binary().filter(is_decodable),
)


@given(st.builds(zip, st.lists(st.text(), unique=True), st.lists(value_strategy)))
def test_randomized_mapping(zipper: Any) -> None:
    """Test randomized mappings"""
    assert is_serializer_compatible(dict(zipper))


@given(st.recursive(value_strategy, lambda trees: st.dictionaries(st.text(), trees)))
def test_randomized_nested_mapping(data: Any) -> None:
    """Test randomized nested mappings"""
    assert is_serializer_compatible({"key": data})


def test_encode_nan() -> None:
    """Test encode Nan."""
    case = {
        "key": float("nan"),
    }
    serialized = serializer.DictProtobufStructSerializer.encode(case)
    deserialized = serializer.DictProtobufStructSerializer.decode(serialized)
    assert math.isnan(deserialized["key"])


@pytest.mark.skip
def test_fuzz_encode() -> None:
    """Fuzz test for serializer. Run directly as a function, not through pytest"""

    @atheris.instrument_func
    def test_encode(input_bytes: bytes) -> None:
        """Test encode decode logic."""
        fdp = atheris.FuzzedDataProvider(input_bytes)
        case = {
            "key1": fdp.ConsumeBool(),
            "key2": fdp.ConsumeFloat(),
            "key3": fdp.ConsumeInt(4),
            "key4": fdp.ConsumeString(12),
            "key5": fdp.ConsumeBytes(12),
            "key6": Struct(),
            "_need_patch": {},
        }
        serializer.to_bytes(case)

    atheris.instrument_all()
    atheris.Setup(sys.argv, test_encode)
    atheris.Fuzz()


def test_encode_non_unicode_raises() -> None:
    """Test encode non unicode."""
    case = {
        "key": b"\xb2\xda\xda\x1a",
    }
    with pytest.raises(UnicodeDecodeError):
        serializer.DictProtobufStructSerializer.encode(case)


def test_sentinel_raises() -> None:
    """Test SENTINEL."""
    case = {
        "SENTINEL": 1,
    }
    with pytest.raises(ValueError):
        serializer.to_bytes(case)



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2021-2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------
"""
Serialize nested dictionaries to bytes using google.protobuf.Struct.

Prerequisites:
- All keys must be of type: str
- Values must be of type: bool, int, float, str, bytes, dict
- Strings must be unicode, as google.protobuf.Struct does not support bytes

The following encoding is required and performed,
and sentinel values are added for decoding:
- bytes to string
- integer to string
"""

import copy
from typing import Any, Dict

from google.protobuf.struct_pb2 import Struct


SENTINEL = "SENTINEL"
ENCODING = "utf-8"


def to_bytes(data: Dict[str, Any]) -> bytes:
    """Serialize to bytes using protobuf. Adds extra data for type-casting."""

    if SENTINEL in data.keys():
        raise ValueError("SENTINEL is a serializer-reserved keyword")
    pstruct = Struct()
    pstruct.update(patch(copy.deepcopy(data)))  # pylint: disable=no-member
    return pstruct.SerializeToString(deterministic=True)


def from_bytes(buffer: bytes) -> Dict[str, Any]:
    """Deserialize patched-up python dict from protobuf bytes."""

    pstruct = Struct()
    pstruct.ParseFromString(buffer)
    return unpatch(dict(pstruct))


def patch(data: Dict[str, Any]) -> Dict[str, Any]:
    """Patch for protobuf serialization. In-place operation."""

    patches: Dict[str, str] = {}
    for key, value in data.items():
        if isinstance(value, (bool, float, str, Struct)):
            pass
        elif isinstance(value, bytes):
            data[key], patches[key] = value.decode(ENCODING), "bytes"
        elif isinstance(value, int):
            data[key], patches[key] = str(value), "int"
        elif isinstance(value, dict):
            data[key], patches[key] = patch(value), "dict"
        else:
            raise NotImplementedError(f"Encoding of `{type(value)}` not supported")

    if patches:
        data[SENTINEL] = patches

    return data


def unpatch(data: Dict[str, Any]) -> Dict[str, Any]:
    """Unpatch for protobuf deserialization. In-place operation."""

    patches = dict(data.pop(SENTINEL, {}))
    for key, value in data.items():
        if isinstance(value, Struct):
            if value == Struct():
                data[key] = (
                    dict() if key in patches and patches[key] == "dict" else Struct()
                )
                continue
            data[key] = unpatch(dict(value))
        elif key in patches:
            data[key] = int(value) if patches[key] == "int" else value.encode(ENCODING)
        elif isinstance(value, (bool, float, str, int)):
            continue
        else:  # pragma: nocover
            raise NotImplementedError(f"Encoding of `{type(value)}` not supported")

    return data


class DictProtobufStructSerializer:  # pylint: disable=too-few-public-methods
    """Class to keep backwards compatibility"""

    encode = to_bytes
    decode = from_bytes



# -*- coding: utf-8 -*-
# ------------------------------------------------------------------------------
#
#   Copyright 2022 Valory AG
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.
#
# ------------------------------------------------------------------------------

"""Develop CLI module."""

import time
from pathlib import Path
from typing import Any, Dict

import click
import yaml

from autonomy.constants import DEFAULT_BUILD_FOLDER
from autonomy.replay.agent import AgentRunner
from autonomy.replay.tendermint import app as proxy_app
from autonomy.replay.tendermint import tendermint_network
from autonomy.replay.utils import fix_address_books, fix_config_files


REGISTRY_PATH = Path("packages/").absolute()
BUILD_DIR = Path(DEFAULT_BUILD_FOLDER).absolute()


@click.group(name="replay")
def replay_group() -> None:
    """Replay tools."""


@replay_group.command(name="agent")
@click.argument("agent", type=int, required=True)
@click.option(
    "--build",
    "build_path",
    type=click.Path(exists=True, dir_okay=True),
    default=BUILD_DIR,
    help="Path to build dir.",
)
@click.option(
    "--registry",
    "registry_path",
    type=click.Path(exists=True, dir_okay=True),
    default=REGISTRY_PATH,
    help="Path to registry folder.",
)
def run_agent(agent: int, build_path: Path, registry_path: Path) -> None:
    """Agent runner."""
    build_path = Path(build_path).absolute()
    registry_path = Path(registry_path).absolute()

    docker_compose_file = build_path / "docker-compose.yaml"
    docker_compose_config = load_docker_config(docker_compose_file)
    agent_data = docker_compose_config["services"][f"abci{agent}"]
    runner = AgentRunner(agent, agent_data, registry_path)
    try:
        runner.start()
        while True:
            time.sleep(1)
    except KeyboardInterrupt:
        runner.stop()


@replay_group.command(name="tendermint")
@click.option(
    "--build",
    "build_dir",
    type=click.Path(dir_okay=True, exists=True),
    default=BUILD_DIR,
    help="Path to build directory.",
)
def run_tendermint(build_dir: Path) -> None:
    """Tendermint runner."""

    build_dir = Path(build_dir).absolute()
    dump_dir = build_dir / "persistent_data" / "dumps"

    fix_address_books(build_dir)
    fix_config_files(build_dir)

    try:
        tendermint_network.init(dump_dir)
    except FileNotFoundError as e:
        raise click.ClickException(str(e)) from e

    try:
        tendermint_network.start()
        proxy_app.run(host="localhost", port=8080)
    except KeyboardInterrupt:
        tendermint_network.stop()


def load_docker_config(file_path: Path) -> Dict[str, Any]:
    """Load docker config."""
    with open(str(file_path), "r", encoding="utf-8") as fp:
        docker_compose_config = yaml.safe_load(fp)

    return docker_compose_config
